'''
JSONRPC Request convenient class helper.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF typing.ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import typing
from collections import OrderedDict
import json
import uuid


class BaseRequestType(type):
    '''
    Base class for both, request and notifications
    '''

    def __getattr__(cls: typing.Callable, name: str) -> typing.Callable:

        def attr_handler(*args: typing.Any, **kwargs: typing.Any) -> "Request":
            return cls(name, *args, **kwargs)

        return attr_handler


class Notification(dict, metaclass=BaseRequestType):
    '''
    Convenient class to create JSONRPC objects(Notifications) and strings without the need to specify redundant information.

    Examples:

        Notification.foo_bar() = > {"jsonrpc": "2.0", "method": "foo_bar"}

        Notification.foo_bar({"hello"="world"}) =>
            {
                    "jsonrpc": "2.0",
                    "method": "foo_bar",
                    "params": {
                        "hello": "world"
                    }
                }

        Notification.foo_bar(var={"hello":"world"}) =>
            {
                "jsonrpc": "2.0",
                "method": "foo_bar",
                "params": {
                    "var": {
                        "hello": "world"
                    }
                }
            }
        Notification.foo_bar(fqdn=["yahoo.com", "trafficserver.org"]) =>
            {
                "jsonrpc": "2.0",
                "method": "foo_bar",
                "params": {
                    "fqdn": ["yahoo.com", "trafficserver.org"]
                }
            }

    '''

    def __init__(self, method: str, *args: typing.Any, **kwargs: typing.Any):
        super(Notification, self).__init__(jsonrpc='2.0', method=method)
        if args and kwargs:
            plist = list(args)
            plist.append(kwargs)
            self.update(params=plist)
        elif args:
            if isinstance(args, tuple):
                # fix this. this is to avoid having params=[[values] which seems invalid. Double check this as [[],[]] would be ok.
                self.update(params=args[0])
            else:
                self.update(params=list(args))
        elif kwargs:
            self.update(params=kwargs)

    # Allow using the dict item as an attribute.
    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def is_notification(self):
        return True

    def __str__(self) -> str:
        return json.dumps(self)


class Request(Notification):
    '''
    Convenient class to create JSONRPC objects and strings without the need to specify redundant information like
    version or the id. All this will be generated automatically.

    Examples:

        Request.foo_bar() = > {"id": "e9ad55fe-d5a6-11eb-a2fd-fa163e6d2ec5", "jsonrpc": "2.0", "method": "foo_bar"}

        Request.foo_bar({"hello"="world"}) =>
            {
                    "id": "850d2998-d5a7-11eb-bebc-fa163e6d2ec5",
                    "jsonrpc": "2.0",
                    "method": "foo_bar",
                    "params": {
                        "hello": "world"
                    }
                }

        Request.foo_bar(var={"hello":"world"}) =>
            {
                "id": "850d2e84-d5a7-11eb-bebc-fa163e6d2ec5",
                "jsonrpc": "2.0",
                "method": "foo_bar",
                "params": {
                    "var": {
                        "hello": "world"
                    }
                }
            }
        Request.foo_bar(fqdn=["yahoo.com", "trafficserver.org"]) =>
            {
                "id": "850d32a8-d5a7-11eb-bebc-fa163e6d2ec5",
                "jsonrpc": "2.0",
                "method": "foo_bar",
                "params": {
                    "fqdn": ["yahoo.com", "trafficserver.org"]
                }
            }

    Note: Use full namespace to avoid name collision => jsonrpc.Request, jsonrpc.Response, etc.
    '''

    def __init__(self, method: str, *args: typing.Any, **kwargs: typing.Any):
        if 'id' in kwargs:
            self.update(id=kwargs.pop('id'))  # avoid duplicated
        else:
            self.update(id=str(uuid.uuid1()))

        super(Request, self).__init__(method, *args, **kwargs)

    def is_notification(self):
        return False


class BatchRequest(list):

    def __init__(self, *args: typing.Union[Request, Notification]):
        for r in args:
            self.append(r)

    def add_request(self, req: typing.Union[Request, Notification]):
        self.append(req)

    def __str__(self) -> str:
        return json.dumps(self)


class Response(dict):
    '''
    Convenient class to help handling jsonrpc responses. This can be source from text directly or by an already parsed test into
    a json object.
    '''

    def __init__(self, *arg, **kwargs):
        if 'text' in kwargs:
            self.__dict__ = json.loads(kwargs['text'])
        elif 'json' in kwargs:
            self.__dict__ = kwargs['json']

    def is_error(self) -> bool:
        '''
        Check whether the error field is present in the response. ie:
            {
                "jsonrpc":"2.0",
                "error":{...},
                "id":"284e0b86-d03a-11eb-9206-fa163e6d2ec5"
            }
        '''
        if 'error' in self.__dict__:
            return True
        return False

    def is_only_success(self) -> bool:
        '''
        Some responses may only have set the result as success. This functions checks that the value in the response field only
        contains the 'success' string, ie:

            {
                "jsonrpc": "2.0",
                "result": "success",
                "id": "8504569c-d5a7-11eb-bebc-fa163e6d2ec5"
            }
        '''
        if self.is_ok() and self.result == 'success':
            return True

        return False

    def is_ok(self) -> bool:
        '''
        No error present in the response, the result field was properly set.
        '''
        return 'result' in self.__dict__

    def error_as_str(self):
        '''
        Build up the error string.
        {
            "jsonrpc":"2.0",
            "error":{
                "code":9,
                "message":"Error during execution",
                "data":[
                    {
                        "code":10001,
                        "message":"No values provided"
                    }
                ]
            },
            "id":"284e0b86-d03a-11eb-9206-fa163e6d2ec5"
            }
        '''
        if self.is_ok():
            return "no error"

        errStr = ""
        errStr += f"\n[code: {self.error['code']}, message: \"{self.error['message']}\"]"
        if 'data' in self.error:
            errStr += "\n\tAdditional Information:"
            for err in self.error['data']:
                errStr += f"\n\t - code: {err['code']}, message: \"{err['message']}\"."
        return errStr

    def __str__(self) -> str:
        return json.dumps(self.__dict__)

    def is_execution_error(self):
        '''
        Checks if the provided error is an Execution error(9).
        '''
        if self.is_ok():
            return False

        return self.error['code'] == 9

    def contains_nested_error(self, code=None, msg=None):
        if self.is_execution_error():
            for err in self.error['data']:
                if code and msg:
                    return err['code'] == code and err['message'] == msg
                elif code and err['code'] == code:
                    return True
                elif msg and err['message'] == msg:
                    return True
                else:
                    return False
        return False


def make_response(text):
    if text == '':
        return None

    s = json.loads(text)
    if isinstance(s, dict):
        return Response(json=s)
    elif isinstance(s, list):
        batch = []
        for r in s:
            batch.append(Response(json=r))
        return batch



#!/usr/bin/env python3
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import sys
import json
import traceback
import yaml
import tempfile
import fileinput
import time
import os.path
import io

from io import StringIO
from string import Template
from pydoc import locate
from jsonschema import validate
from jsonschema.exceptions import ValidationError
from colorama import init, Fore, Back, Style

init(autoreset=True)

Remapped_Vars = {}

Renamed_Records = {
    # as it can have a numeric value [0-2]
    'proxy.config.output.logfile': 'proxy.config.output.logfile.name',
    'proxy.config.exec_thread.autoconfig': 'proxy.config.exec_thread.autoconfig.enabled',
    'proxy.config.hostdb': 'proxy.config.hostdb.enabled',
    'proxy.config.tunnel.prewarm': 'proxy.config.tunnel.prewarm.enabled',
    'proxy.config.ssl.origin_session_cache': 'proxy.config.ssl.origin_session_cache.enabled',
    'proxy.config.ssl.session_cache': 'proxy.config.ssl.session_cache.value',
    'proxy.config.ssl.TLSv1_3': 'proxy.config.ssl.TLSv1_3.enabled',
    'proxy.config.ssl.client.TLSv1_3': 'proxy.config.ssl.client.TLSv1_3.enabled'
}

###############################################################################################
###############################################################################################


def red(text, bright=False):
    return f"{Style.BRIGHT if bright else ''}{Fore.RED}{text}{Style.RESET_ALL}"


def bright(text, color=Fore.GREEN):
    return f'{Style.BRIGHT}{color}{text}{Style.RESET_ALL}'


def error():
    return f"{red('[E] ', True)}"


def fancy_chain_print(lst, isError=False):
    output = StringIO()
    if len(lst) == 1:
        output.write(f'» {lst[0]}\n')
        s = output.getvalue()
        output.close()
        return s
    space = ''
    output.write(f'┌■ {lst[0]}\n')
    if len(lst) > 2:
        space = ' '
        # from 2nd till end-1
        output.write(f'└┬──» {lst[1]}\n')
        for x in lst[2:-1]:
            output.write(f' ├──» {x}\n')

    z = lst[-1]
    output.write(f'{space}└──» {z}\n')

    s = output.getvalue()
    output.close()
    return s


def progressbar(it, size=60, out=sys.stdout, mute=False):
    count = len(it)

    def show(j):
        x = int(size * j / count)
        if not mute:
            print("[{}{}] {}/{}".format("█" * x, "░" * (size - x), j, count), end='\r', file=out, flush=True)

    show(0)
    for i, item in enumerate(it):
        yield item

        show(i + 1)
    if not mute:
        print("\n", flush=True, file=out)


def validate_schema(data, schema_filename):
    if schema_filename:
        with open(schema_filename, 'r') as f:
            s = f.read()
            sdata = json.loads(s)
    try:
        # validate will throw if invalid schema.
        if schema_filename:
            validate(instance=data, schema=sdata)
    except ValidationError as ve:
        return str(ve)

    return ""


def save_to_file(filename, is_json, typerepr, no_backup, data):

    def float_representer(dumper, value):
        return dumper.represent_scalar(u'tag:yaml.org,2002:float', str(value), style="'")

    def int_representer(dumper, value):
        return dumper.represent_scalar(u'tag:yaml.org,2002:int', str(value), style="'")

    def bool_representer(dumper, value):
        return dumper.represent_scalar(u'tag:yaml.org,2002:bool', str(value), style="'")

    def null_representer(dumper, value):
        return dumper.represent_scalar(u'tag:yaml.org,2002:null', str(value), style="'")

    # Make sure we do not wipe out an existing file.
    if no_backup == False and os.path.exists(filename):
        os.system(f'cp {filename} {filename}.{int(time.time())}')

    with open(filename, 'w') as f:
        if is_json:
            json.dump(data, f, indent=4, sort_keys=True)
        else:
            if 'float' in typerepr:
                yaml.add_representer(float, float_representer)
            if 'int' in typerepr:
                yaml.add_representer(int, int_representer)
            if 'bool' in typerepr:
                yaml.add_representer(bool, bool_representer)
            if 'null' in typerepr:
                yaml.add_representer(None, null_representer)
            yaml.dump(data, f)


def print_summary():
    if len(Remapped_Vars):
        out = []
        out.append(f'{Style.BRIGHT}{Fore.GREEN}{len(Remapped_Vars)} Renamed records:{Style.RESET_ALL}')
        idx = 1
        for k, v in Remapped_Vars.items():
            out.append(f'{Style.BRIGHT}{Fore.GREEN}#{idx}{Style.RESET_ALL} : {k} -> {Style.BRIGHT}{Fore.GREEN}{v}{Style.RESET_ALL}')
            idx = idx + 1

        print(fancy_chain_print(out))


def get_value(type, value):

    def have_multipliers(value, mps):
        for m in mps:
            if value.endswith(m):
                return True

        return False

    if value == 'nullptr' or value == 'NULL':
        return None

    # We want to make the type right when inserting the element into the dict.
    if type == 'FLOAT':
        return float(value)
    elif type == 'INT':
        # We need to make it YAML compliant as this will be an int, so if contains
        # any special character like hex or a multiplier, then we make a string. ATS will
        # parse it as string anyway.
        if value.startswith('0x') or have_multipliers(value, ['K', 'M', 'G', 'T']):
            return str(value)
        else:
            return int(value)
    elif type == 'STRING':
        return str(value)

    return None


def add_object(config, var, value, type, track_info):
    key = ''
    index = var.find('.')
    if index < 0:  # last part
        if isinstance(config, dict) == False:
            line, rec = track_info
            raise Exception(
                f"We cannot continue with '{rec}' at line '{line}' as a value node will be overridden.\nPlease check your config.")
        elif var in config and isinstance(config[var], dict):
            line, rec = track_info
            raise Exception(
                f"We cannot continue with '{rec}' at line '{line}' as an existing YAML map will be overridden.\nPlease check your config."
            )
        config[var] = get_value(type, value)
    else:
        key = var[:index]
        if key not in config:
            config[key] = {}

        add_object(config[key], var[index + 1:], value, type, track_info)


def fix_record_names(file):
    temp = tempfile.NamedTemporaryFile(prefix='records.config_', delete=False)

    for l in fileinput.input(files=file):
        elements = l.split()
        if len(elements) == 4:
            rec_name = elements[1]
            if rec_name in Renamed_Records:
                elements[1] = Renamed_Records[rec_name]

                rec_line = ' '.join(elements)
                temp.write(rec_line.encode('utf-8'))
                temp.write(b'\n')
                Remapped_Vars[rec_name] = Renamed_Records[rec_name]

                continue
        # if elements len diff from 4 .. well.. it shouldn't.
        temp.write(l.encode('utf-8'))

    temp.close()
    return temp.name


def handle_file_input(args):
    config = {}

    # Fix the names first.
    filename = fix_record_names(args.file)
    with open(filename, "r+") as f:
        lines = f.readlines()
        num_lines = len(lines)
        for idx in progressbar(range(num_lines), 40, mute=args.mute):
            line = lines[idx]
            if '#' == line[0] or '\n' == line[0] or ' ' == line[0]:
                idx = idx + 1
                continue
            s = line.split(' ', maxsplit=3)  # keep values all together
            name = s[1]
            type = s[2]
            value = s[3]

            track_info = (idx + 1, name)  # in case we want to show any error. rec name is always handy.
            # We ignore the prefix and work away from  there.
            if name.startswith("proxy.config."):
                name = name[len("proxy.config."):]
            elif name.startswith("local.config."):
                name = name[len("local.config."):]
            elif args.node and name.startswith("proxy.node."):
                name = name[len("proxy."):]

            # Build the object
            add_object(config, name, value[:-1], type, track_info)
            idx = idx + 1

    ts = {}
    ts['records'] = config

    if args.schema:
        err = validate_schema(ts, args.schema)
        if err:
            raise Exception("Schema failed.")

    save_to_file(args.output, args.json, args.typerepr, args.no_backup, ts)

    f.close()
    return


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='records.config to YAML/JSON convert tool')
    parser.add_argument('-f', '--file', help='records.config input file.', required=False, default="records.config")
    parser.add_argument('-n', '--node', help="Include 'proxy.node' variables in the parser.", action='store_true')
    parser.add_argument('-t', '--typerepr', help="Use type representer (list)", required=False, default=[''])
    parser.add_argument('-s', '--schema', help="Validate the output using a json schema file.")
    parser.add_argument('-o', '--output', help="Save to output file.", required=False, default="records.yaml")
    parser.add_argument(
        '-m', '--mute', help="Be quiet, do not output anything, except for errors", required=False, action='store_true')
    parser.add_argument('-e', '--error', help="Show traceback", required=False, action='store_true', default=False)
    parser.add_argument(
        '-x',
        '--no-backup',
        help="If it exists, do not backup the output file.",
        required=False,
        action='store_true',
        default=False,
        dest='no_backup')
    kk = parser.add_mutually_exclusive_group(required=False)
    kk.add_argument('-j', '--json', help="Output as json", action='store_true', default=False)
    kk.add_argument('-y', '--yaml', help="Output as yaml", action='store_true', default=True)
    parser.set_defaults(func=handle_file_input)
    args = parser.parse_args()

    try:
        func = args.func
    except AttributeError:
        parser.error("Too few arguments")
    try:
        func(args)
        if args.mute is False:
            print_summary()

    except Exception as e:
        print("Something went wrong:\n{}".format(e))

        if args.error:
            print(f"\n{traceback.format_exc()}")
        sys.exit(1)

    sys.exit(0)



'''
Test the Forwarded header and related configuration..
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Test FORWARDED header.
'''

Test.SkipUnless(
    Condition.HasCurlFeature('http2'),
    Condition.HasCurlFeature('IPv6'),
)
Test.ContinueOnFail = True

testName = "FORWARDED"

server = Test.MakeOriginServer("server", options={'--load': os.path.join(Test.TestDirectory, 'forwarded-observer.py')})

request_header = {"headers": "GET / HTTP/1.1\r\nHost: www.no-oride.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
response_header = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", request_header, response_header)
request_header = {"headers": "GET / HTTP/1.1\r\nHost: www.forwarded-none.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", request_header, response_header)
request_header = {"headers": "GET / HTTP/1.1\r\nHost: www.forwarded-for.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", request_header, response_header)
request_header = {"headers": "GET / HTTP/1.1\r\nHost: www.forwarded-by-ip.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", request_header, response_header)
request_header = {
    "headers": "GET / HTTP/1.1\r\nHost: www.forwarded-by-unknown.com\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
server.addResponse("sessionlog.json", request_header, response_header)
request_header = {
    "headers": "GET / HTTP/1.1\r\nHost: www.forwarded-by-server-name.com\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
server.addResponse("sessionlog.json", request_header, response_header)
request_header = {"headers": "GET / HTTP/1.1\r\nHost: www.forwarded-by-uuid.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", request_header, response_header)
request_header = {"headers": "GET / HTTP/1.1\r\nHost: www.forwarded-proto.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", request_header, response_header)
request_header = {"headers": "GET / HTTP/1.1\r\nHost: www.forwarded-host.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", request_header, response_header)
request_header = {
    "headers": "GET / HTTP/1.1\r\nHost: www.forwarded-connection-compact.com\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
server.addResponse("sessionlog.json", request_header, response_header)
request_header = {
    "headers": "GET / HTTP/1.1\r\nHost: www.forwarded-connection-std.com\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
server.addResponse("sessionlog.json", request_header, response_header)
request_header = {
    "headers": "GET / HTTP/1.1\r\nHost: www.forwarded-connection-full.com\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
server.addResponse("sessionlog.json", request_header, response_header)

# Set up to check the output after the tests have run.
#
forwarded_log_id = Test.Disk.File("forwarded.log")
forwarded_log_id.Content = "forwarded.gold"


def baselineTsSetup(ts):

    ts.addDefaultSSLFiles()

    ts.Disk.records_config.update(
        {
            # 'proxy.config.diags.debug.enabled': 1,
            'proxy.config.url_remap.pristine_host_hdr': 1,  # Retain Host header in original incoming client request.
            'proxy.config.proxy_name': 'Poxy_Proxy',  # This will be the server name.
            'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
            'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir)
        })

    ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

    ts.Disk.remap_config.AddLine('map http://www.no-oride.com http://127.0.0.1:{0}'.format(server.Variables.Port))


# Disable the cache to make sure each request is forwarded to the origin
# server.
ts = Test.MakeATSProcess("ts", enable_tls=True, enable_cache=False)

baselineTsSetup(ts)

ts.Disk.remap_config.AddLine(
    'map http://www.forwarded-none.com http://127.0.0.1:{0}'.format(server.Variables.Port) +
    ' @plugin=conf_remap.so @pparam=proxy.config.http.insert_forwarded=none')
ts.Disk.remap_config.AddLine(
    'map http://www.forwarded-for.com http://127.0.0.1:{0}'.format(server.Variables.Port) +
    ' @plugin=conf_remap.so @pparam=proxy.config.http.insert_forwarded=for')
ts.Disk.remap_config.AddLine(
    'map http://www.forwarded-by-ip.com http://127.0.0.1:{0}'.format(server.Variables.Port) +
    ' @plugin=conf_remap.so @pparam=proxy.config.http.insert_forwarded=by=ip')
ts.Disk.remap_config.AddLine(
    'map http://www.forwarded-by-unknown.com http://127.0.0.1:{0}'.format(server.Variables.Port) +
    ' @plugin=conf_remap.so @pparam=proxy.config.http.insert_forwarded=by=unknown')
ts.Disk.remap_config.AddLine(
    'map http://www.forwarded-by-server-name.com http://127.0.0.1:{0}'.format(server.Variables.Port) +
    ' @plugin=conf_remap.so @pparam=proxy.config.http.insert_forwarded=by=serverName')
ts.Disk.remap_config.AddLine(
    'map http://www.forwarded-by-uuid.com http://127.0.0.1:{0}'.format(server.Variables.Port) +
    ' @plugin=conf_remap.so @pparam=proxy.config.http.insert_forwarded=by=uuid')
ts.Disk.remap_config.AddLine(
    'map http://www.forwarded-proto.com http://127.0.0.1:{0}'.format(server.Variables.Port) +
    ' @plugin=conf_remap.so @pparam=proxy.config.http.insert_forwarded=proto')
ts.Disk.remap_config.AddLine(
    'map http://www.forwarded-host.com http://127.0.0.1:{0}'.format(server.Variables.Port) +
    ' @plugin=conf_remap.so @pparam=proxy.config.http.insert_forwarded=host')
ts.Disk.remap_config.AddLine(
    'map http://www.forwarded-connection-compact.com http://127.0.0.1:{0}'.format(server.Variables.Port) +
    ' @plugin=conf_remap.so @pparam=proxy.config.http.insert_forwarded=connection=compact')
ts.Disk.remap_config.AddLine(
    'map http://www.forwarded-connection-std.com http://127.0.0.1:{0}'.format(server.Variables.Port) +
    ' @plugin=conf_remap.so @pparam=proxy.config.http.insert_forwarded=connection=std')
ts.Disk.remap_config.AddLine(
    'map http://www.forwarded-connection-full.com http://127.0.0.1:{0}'.format(server.Variables.Port) +
    ' @plugin=conf_remap.so @pparam=proxy.config.http.insert_forwarded=connection=full')

# Basic HTTP 1.1 -- No Forwarded by default
tr = Test.AddTestRun()
# Wait for the micro server
tr.Processes.Default.StartBefore(server, ready=When.PortOpen(server.Variables.Port))
# Delay on readiness of our ssl ports
tr.Processes.Default.StartBefore(Test.Processes.ts)
#
tr.Processes.Default.Command = (
    'curl --verbose --ipv4 --http1.1 --proxy localhost:{} http://www.no-oride.com'.format(ts.Variables.port))
tr.Processes.Default.ReturnCode = 0


def TestHttp1_1(host):

    tr = Test.AddTestRun()
    tr.Processes.Default.Command = (
        'curl --verbose --ipv4 --http1.1 --proxy localhost:{} http://{}'.format(ts.Variables.port, host))
    tr.Processes.Default.ReturnCode = 0


# Basic HTTP 1.1 -- No Forwarded -- explicit configuration.
#
TestHttp1_1('www.forwarded-none.com')

# Test enabling of each forwarded parameter singly.

TestHttp1_1('www.forwarded-for.com')

# Note:  forwarded-observer.py counts on the "by" tests being done in the order below.

TestHttp1_1('www.forwarded-by-ip.com')
TestHttp1_1('www.forwarded-by-unknown.com')
TestHttp1_1('www.forwarded-by-server-name.com')
TestHttp1_1('www.forwarded-by-uuid.com')

TestHttp1_1('www.forwarded-proto.com')
TestHttp1_1('www.forwarded-host.com')
TestHttp1_1('www.forwarded-connection-compact.com')
TestHttp1_1('www.forwarded-connection-std.com')
TestHttp1_1('www.forwarded-connection-full.com')

ts2 = Test.MakeATSProcess("ts2", enable_tls=True)

baselineTsSetup(ts2)

ts2.Disk.records_config.update(
    {
        # 'proxy.config.diags.debug.enabled': 1,
        'proxy.config.url_remap.pristine_host_hdr': 1,  # Retain Host header in original incoming client request.
        'proxy.config.http.insert_forwarded': 'by=uuid'
    })

ts2.Disk.remap_config.AddLine('map https://www.no-oride.com http://127.0.0.1:{0}'.format(server.Variables.Port))

# Forwarded header with UUID of 2nd ATS.
tr = Test.AddTestRun()
# Delay on readiness of our ssl ports
tr.Processes.Default.StartBefore(Test.Processes.ts2)
#
tr.Processes.Default.Command = (
    'curl --verbose --ipv4 --http1.1 --proxy localhost:{} http://www.no-oride.com'.format(ts2.Variables.port))
tr.Processes.Default.ReturnCode = 0

# Call traffic_ctrl to set insert_forwarded
tr = Test.AddTestRun()
tr.Processes.Default.Command = (
    'traffic_ctl --debug config set proxy.config.http.insert_forwarded' +
    ' "for|by=ip|by=unknown|by=servername|by=uuid|proto|host|connection=compact|connection=std|connection=full"')
tr.Processes.Default.ForceUseShell = False
tr.Processes.Default.Env = ts2.Env
tr.Processes.Default.ReturnCode = 0

# HTTP 1.1
tr = Test.AddTestRun()
# Delay to give traffic_ctl config change time to take effect.
tr.DelayStart = 15
tr.Processes.Default.Command = (
    'curl --verbose --ipv4 --http1.1 --proxy localhost:{} http://www.no-oride.com'.format(ts2.Variables.port))
tr.Processes.Default.ReturnCode = 0

# HTTP 1.0
tr = Test.AddTestRun()
tr.Processes.Default.Command = (
    'curl --verbose --ipv4 --http1.0 --proxy localhost:{} http://www.no-oride.com'.format(ts2.Variables.port))
tr.Processes.Default.ReturnCode = 0

# HTTP 1.0 -- Forwarded headers already present
tr = Test.AddTestRun()
tr.Processes.Default.Command = (
    "curl --verbose -H 'forwarded:for=0.6.6.6' -H 'forwarded:for=_argh' --ipv4 --http1.0" +
    " --proxy localhost:{} http://www.no-oride.com".format(ts2.Variables.port))
tr.Processes.Default.ReturnCode = 0

# HTTP 2
tr = Test.AddTestRun()
tr.Processes.Default.Command = (
    'curl --verbose --ipv4 --http2 --insecure --header "Host: www.no-oride.com"' +
    ' https://localhost:{}'.format(ts2.Variables.ssl_port))
tr.Processes.Default.ReturnCode = 0

# TLS
tr = Test.AddTestRun()
tr.Processes.Default.Command = (
    'curl --verbose --ipv4 --http1.1 --insecure --header "Host: www.no-oride.com" https://localhost:{}'.format(
        ts2.Variables.ssl_port))
tr.Processes.Default.ReturnCode = 0

# IPv6

tr = Test.AddTestRun()
tr.Processes.Default.Command = (
    'curl --verbose --ipv6 --http1.1 --proxy localhost:{} http://www.no-oride.com'.format(ts2.Variables.portv6))
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun()
tr.Processes.Default.Command = (
    'curl --verbose --ipv6 --http1.1 --insecure --header "Host: www.no-oride.com" https://localhost:{}'.format(
        ts2.Variables.ssl_portv6))
tr.Processes.Default.ReturnCode = 0



'''
Test proxy serving stale content when DNS lookup fails
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.ContinueOnFail = True
# Set up hierarchical caching processes
ts_child = Test.MakeATSProcess("ts_child")
ts_parent = Test.MakeATSProcess("ts_parent")
nameserver = Test.MakeDNServer("dns")
server_name = "http://unknown.domain.com/"

Test.testName = "STALE"

# Config child proxy to route to parent proxy
ts_child.Disk.records_config.update(
    {
        'proxy.config.http.push_method_enabled': 1,
        'proxy.config.url_remap.pristine_host_hdr': 1,
        'proxy.config.http.cache.max_stale_age': 10,
        'proxy.config.http.parent_proxy.self_detect': 0,
        'proxy.config.dns.nameservers': f"127.0.0.1:{nameserver.Variables.Port}",
    })
ts_child.Disk.parent_config.AddLine(
    f'dest_domain=. parent=localhost:{ts_parent.Variables.port} round_robin=consistent_hash go_direct=false')
ts_child.Disk.remap_config.AddLine(f'map http://localhost:{ts_child.Variables.port} {server_name}')

# Configure parent proxy
ts_parent.Disk.records_config.update(
    {
        'proxy.config.http.push_method_enabled': 1,
        'proxy.config.url_remap.pristine_host_hdr': 1,
        'proxy.config.http.cache.max_stale_age': 10,
        'proxy.config.dns.nameservers': f"127.0.0.1:{nameserver.Variables.Port}",
    })
ts_parent.Disk.remap_config.AddLine(f'map http://localhost:{ts_parent.Variables.port} {server_name}')
ts_parent.Disk.remap_config.AddLine(f'map {server_name} {server_name}')

# Configure nameserver
nameserver.addRecords(records={"localhost": ["127.0.0.1"]})

# Object to push to proxies
stale_5 = "HTTP/1.1 200 OK\nServer: ATS/10.0.0\nAccept-Ranges: bytes\nContent-Length: 6\nCache-Control: public, max-age=5\n\nCACHED"
stale_10 = "HTTP/1.1 200 OK\nServer: ATS/10.0.0\nAccept-Ranges: bytes\nContent-Length: 6\nCache-Control: public, max-age=10\n\nCACHED"

# Testing scenarios
child_curl_request = (
    # Test child serving stale with failed DNS OS lookup
    f'curl -X PUSH -d "{stale_5}" "http://localhost:{ts_child.Variables.port}";'
    f'curl -X PUSH -d "{stale_10}" "http://localhost:{ts_parent.Variables.port}";'
    f'sleep 7; curl -s -v http://localhost:{ts_child.Variables.port};'
    f'sleep 15; curl -s -v http://localhost:{ts_child.Variables.port};'
    # Test parent serving stale with failed DNS OS lookup
    f'curl -X PUSH -d "{stale_5}" "http://localhost:{ts_parent.Variables.port}";'
    f'sleep 7; curl -s -v http://localhost:{ts_parent.Variables.port};'
    f'sleep 15; curl -s -v http://localhost:{ts_parent.Variables.port};')

# Test case for when parent server is down but child proxy can serve cache object
tr = Test.AddTestRun()
tr.Processes.Default.Command = child_curl_request
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(ts_child)
tr.Processes.Default.StartBefore(ts_parent)
tr.Processes.Default.StartBefore(nameserver)
tr.Processes.Default.Streams.stderr = "gold/serve_stale_dns_fail.gold"
tr.StillRunningAfter = ts_child
tr.StillRunningAfter = ts_parent



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import subprocess

Test.Summary = '''
Test transactions and sessions, making sure they open and close in the proper order.
'''

# Define default ATS. Disable the cache to simplify the test.
ts = Test.MakeATSProcess("ts", enable_cache=False)

server = Test.MakeOriginServer("server")
server2 = Test.MakeOriginServer("server2")

Test.testName = ""
request_header = {"headers": "GET / HTTP/1.1\r\nHost: oc.test\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
# expected response from the origin server
response_header = {
    "headers": "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Length:0\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}

Test.PrepareTestPlugin(os.path.join(Test.Variables.AtsTestPluginsDir, 'ssntxnorder_verify.so'), ts)

# add response to the server dictionary
server.addResponse("sessionfile.log", request_header, response_header)
ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 0,
        'proxy.config.diags.debug.tags': 'ssntxnorder_verify.*',
        'proxy.config.cache.enable_read_while_writer': 0
    })

ts.Disk.remap_config.AddLine('map http://oc.test:{0} http://127.0.0.1:{1}'.format(ts.Variables.port, server.Variables.Port))

# Add connection close to ensure that the client connection closes promptly after completing the transaction
cmd = 'curl -H "Connection: close" -vs -H "host:oc.test" http://127.0.0.1:{0}'.format(ts.Variables.port)
numberOfRequests = 100

tr = Test.AddTestRun()
# Create a bunch of curl commands to be executed in parallel. Default.Process is set in SpawnCommands.
# On Fedora 28/29, it seems that curl will occasionally timeout after a couple seconds and return exitcode 2
# Examining the packet capture shows that Traffic Server dutifully sends the response
ps = tr.SpawnCommands(cmdstr=cmd, count=numberOfRequests, retcode=Any(0, 2))
tr.Processes.Default.Env = ts.Env
tr.Processes.Default.ReturnCode = Any(0, 2)

# Execution order is: ts/server, ps(curl cmds), Default Process.
tr.Processes.Default.StartBefore(server, ready=When.PortOpen(server.Variables.Port))
tr.Processes.Default.StartBefore(Test.Processes.ts)
ts.StartAfter(*ps)
server.StartAfter(*ps)
tr.StillRunningAfter = ts

# Signal that all the curl processes have completed
tr = Test.AddTestRun("Curl Done")
tr.DelayStart = 2  # Delaying a couple seconds to make sure the global continuation's lock contention resolves.
tr.Processes.Default.Command = "traffic_ctl plugin msg done done"
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Env = ts.Env
tr.StillRunningAfter = ts

# Parking this as a ready tester on a meaningless process
# To stall the test runs that check for the stats until the
# stats have propagated and are ready to read.


def make_done_stat_ready(tsenv):

    def done_stat_ready(process, hasRunFor, **kw):
        retval = subprocess.run(
            "traffic_ctl metric get ssntxnorder_verify.test.done",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env=tsenv)
        return retval.returncode == 0 and b'1' in retval.stdout

    return done_stat_ready


# number of sessions/transactions opened and closed are equal
tr = Test.AddTestRun("Check Ssn order errors")
server2.StartupTimeout = 60
# Again, here the important thing is the ready function not the server2 process
tr.Processes.Default.StartBefore(server2, ready=make_done_stat_ready(ts.Env))
tr.Processes.Default.Command = 'traffic_ctl metric get ssntxnorder_verify.err'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Env = ts.Env
tr.Processes.Default.Streams.All = Testers.ContainsExpression(
    'ssntxnorder_verify.err 0', 'incorrect statistic return, or possible error.')
tr.StillRunningAfter = ts
tr.StillRunningAfter = server

comparator_command = '''
if test "`traffic_ctl metric get ssntxnorder_verify.{0}.start | cut -d ' ' -f 2`" -eq "`traffic_ctl metric get ssntxnorder_verify.{0}.close | cut -d ' ' -f 2`" ; then\
     echo yes;\
    else \
    echo no; \
    fi; \
    traffic_ctl metric match ssntxnorder_verify
    '''

# number of sessions/transactions opened and closed are equal
tr = Test.AddTestRun("Check for ssn open/close")
tr.Processes.Default.Command = comparator_command.format('ssn')
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Env = ts.Env
tr.Processes.Default.Streams.stdout = Testers.ContainsExpression("yes", 'should verify contents')
tr.Processes.Default.Streams.stdout += Testers.ExcludesExpression("ssntxnorder_verify.ssn.start 0", 'should be nonzero')
tr.StillRunningAfter = ts
tr.StillRunningAfter = server

tr = Test.AddTestRun("Check for txn/open/close")
tr.Processes.Default.Command = comparator_command.format('txn')
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Env = ts.Env
tr.Processes.Default.Streams.stdout = Testers.ContainsExpression("yes", 'should verify contents')
tr.Processes.Default.Streams.stdout += Testers.ExcludesExpression("ssntxnorder_verify.txn.start 0", 'should be nonzero')
# and we receive the same number of transactions as we asked it to make
tr.Processes.Default.Streams.stdout += Testers.ContainsExpression(
    "ssntxnorder_verify.txn.start {}".format(numberOfRequests), 'should be the number of transactions we made')
tr.StillRunningAfter = ts
tr.StillRunningAfter = server



'''
Verify ATS handles down name servers correctly.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ports import get_port

# This value tracks DNS_PRIMARY_RETRY_PERIOD in P_DNSProcessor.h.
DNS_PRIMARY_RETRY_PERIOD = 5

Test.Summary = '''
Verify ATS handles down name servers correctly.
'''


class DownDNSNameserverTest:
    """Encapsulate logic for testing down name server functionality.

    Note that this test verifies that ATS behaves correctly with respect to
    down *name* servers. This is different than testing down origin servers,
    which is an entirely different feature and implementation.
    """

    _replay_file = "replay/multiple_host_requests.replay.yaml"

    def __init__(self):
        """Initialize the Test processes for the test runs."""
        self._dns_port = None
        self._configure_origin_server()
        self._configure_traffic_server()
        self._servers_are_running = False

        self._client_counter = 0
        self._dns_counter = 0

    def _configure_origin_server(self):
        """Configure the origin server and the DNS port."""
        self._server = Test.MakeVerifierServerProcess("server", self._replay_file)

        # Associate the DNS port with the long-running server. This way
        # ATS can be configured to use it across the multiple test runs
        # and DNS servers.
        self._dns_port = get_port(self._server, "DNSPort")

    def _configure_traffic_server(self):
        """Configure Traffic Server."""
        if not self._dns_port:
            raise RuntimeError("The DNS port was not configured.")

        # Since we are testing DNS name resolution and whether these
        # transactions reach the origin, disable the caching to ensure none of
        # the responses come out of the cache without going to the origin.
        self._ts = Test.MakeATSProcess("ts", enable_cache=False)

        self._ts.Disk.records_config.update(
            {
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'hostdb|dns',
                'proxy.config.dns.nameservers': f'127.0.0.1:{self._dns_port}',
                'proxy.config.dns.resolv_conf': 'NULL'
            })

        # Cause a name resolution for each, unique path.
        self._ts.Disk.remap_config.AddLines(
            [
                f'map /first/host http://first.host.com:{self._server.Variables.http_port}/',
                f'map /second/host http://second.host.com:{self._server.Variables.http_port}/',
                f'map /third/host http://third.host.com:{self._server.Variables.http_port}/',
            ])

    def _run_transaction(self, start_dns: bool, keyname: str):
        """Run a transaction with the name server reachable.

        :param start_dns: Whether the TestRun should configure a name server to
        be running for ATS to talk to.

        :param keyname: The identifier for the transaction to run.
        """
        tr = Test.AddTestRun()

        tr.AddVerifierClientProcess(
            f'client{self._client_counter}',
            self._replay_file,
            http_ports=[self._ts.Variables.port],
            other_args=f'--keys {keyname}')
        self._client_counter += 1

        if start_dns:
            dns = tr.MakeDNServer(f'dns{self._dns_counter}', default='127.0.0.1', port=self._dns_port)
            self._dns_counter += 1
            tr.Processes.Default.StartBefore(dns)

        if not self._servers_are_running:
            tr.Processes.Default.StartBefore(self._server)
            tr.Processes.Default.StartBefore(self._ts)
            self._servers_are_running = True

        # Verify that the client tried to send the transaction.
        tr.Processes.Default.Streams.All += Testers.ContainsExpression(
            f'uuid: {keyname}', f'The client should have sent a transaction with uuid {keyname}')

        # The client will report an error if ATS could not complete the
        # transaction due to DNS resolution issues.
        if not start_dns:
            tr.Processes.Default.ReturnCode = 1

    def _delay_for_dns_retry_period(self, start_dns: bool):
        """Wait for the failed DNS retry period."""
        tr = Test.AddTestRun()

        if start_dns:
            dns = tr.MakeDNServer(f'dns{self._dns_counter}', default='127.0.0.1', port=self._dns_port)
            self._dns_counter += 1
            tr.Processes.Default.StartBefore(dns)

        # Exceed the retry period.
        tr.Processes.Default.Command = f'sleep {DNS_PRIMARY_RETRY_PERIOD + 1}'

    def _test_dns_reachable(self):
        """This is the base case: DNS is reachable.

        This verifies that ATS can use the name server to resolve a name when
        the name server is reachable.
        """
        self._run_transaction(start_dns=True, keyname='first_host')

    def _test_dns_reachable_within_the_retry_period(self):
        """Test nameserver connectivity resolution within the retry period.

        Have the name server be inaccessible, but bring it back online within
        the retry period.
        """
        # Trigger a name resolution with the name server down.
        self._run_transaction(start_dns=False, keyname='second_host')

        # Wait long enough that ATS to retry the name server, with the server
        # up.
        self._delay_for_dns_retry_period(start_dns=True)

        # ATS should now use the working name server again to resolve the host
        # name.
        self._run_transaction(start_dns=True, keyname='second_host')

    def _test_dns_reachable_outside_the_retry_period(self):
        """Test nameserver connectivity resolution outside the retry period.

        Have the name server be inaccessible, but bring it back online outside
        the retry period. This verifies that we handle it coming back online
        after the first attempt to re-connect to it.
        """
        # Trigger a name resolution with the name server down.
        self._run_transaction(start_dns=False, keyname='third_host')

        # Wait long enough that ATS to retry connectivity to the name server,
        # with the server down. This will naturally fail.
        self._delay_for_dns_retry_period(start_dns=False)

        # Wait long enough that ATS to retry connectivity a second time to name
        # server, this time with the server up.
        self._delay_for_dns_retry_period(start_dns=True)

        # ATS should now try the name server again, detect that it is up, and
        # complete the transaction.
        self._run_transaction(start_dns=True, keyname='third_host')

    def run(self):
        """Exercise ATS down name server functionality."""
        self._test_dns_reachable()
        self._test_dns_reachable_within_the_retry_period()
        self._test_dns_reachable_outside_the_retry_period()


DownDNSNameserverTest().run()



'''
Common utilities for the Proxy Verifier extensions.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os
import tempfile
from string import Template


def create_address_argument(ports):
    """
    >>> create_address_argument([8080, 8081])
    '"127.0.0.1:8080,127.0.0.1:8081"'
    """
    is_first = True
    argument = '"'
    for port in ports:
        if is_first:
            is_first = False
        else:
            argument += ','
        argument += "127.0.0.1:{}".format(port)
    argument += '"'
    return argument


def substitute_context_in_replay_file(process, replay_path, context):
    '''
    Perform substitution base on the passed context dict.
    This function will return the new replay_path file
    '''
    # Only files for now
    if os.path.isdir(replay_path):
        raise ValueError(f"Mapping substitution not supported for directories.")

    with open(os.path.join(process.TestDirectory, replay_path), 'r') as replay_file:
        replay_template = Template(replay_file.read())
        replay_content = replay_template.substitute(context)
        tf = tempfile.NamedTemporaryFile(delete=False, dir=process.RunDirectory, suffix=f"_{os.path.basename(replay_path)}")
        replay_path = tf.name
        with open(replay_path, "w") as new_replay_file:
            new_replay_file.write(replay_content)

    # use this as replay_path
    return replay_path



'''
Test authorization-related caching behaviors
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Test authorization-related caching behaviors
'''

Test.ContinueOnFail = True

# **testname is required**
testName = ""


class AuthDefaultTest:
    # Verify the proper caching behavior for request/response containing
    # auth-related fields when ATS is in default configuration
    authDefaultReplayFile = "replay/auth-default.replay.yaml"

    def __init__(self):
        self.setupOriginServer()
        self.setupTS()

    def setupOriginServer(self):
        self.server = Test.MakeVerifierServerProcess("auth-default-verifier-server", self.authDefaultReplayFile)

    def setupTS(self):
        self.ts = Test.MakeATSProcess("ts-auth-default")
        self.ts.Disk.records_config.update({
            "proxy.config.diags.debug.enabled": 1,
            "proxy.config.diags.debug.tags": "http",
        })
        self.ts.Disk.remap_config.AddLine(f"map / http://127.0.0.1:{self.server.Variables.http_port}/",)

        # Verify log for skipping the WWW-Authenticate response
        self.ts.Disk.traffic_out.Content += Testers.ContainsExpression(
            "response has WWW-Authenticate, response is not cacheable",
            "Verify ATS doesn't store the response with WWW-Authenticate.")

    def runTraffic(self):
        tr = Test.AddTestRun(
            "Verify the proper caching behavior for request/response containing auth-related fields when ATS is in default configuration"
        )
        tr.AddVerifierClientProcess(
            "auth-default-client", self.authDefaultReplayFile, http_ports=[self.ts.Variables.port], other_args='--thread-limit 1')
        tr.Processes.Default.StartBefore(self.server)
        tr.Processes.Default.StartBefore(self.ts)
        tr.StillRunningAfter = self.server
        tr.StillRunningAfter = self.ts

    def run(self):
        self.runTraffic()


AuthDefaultTest().run()


class AuthIgnoredTest:
    # Verify the proper caching behavior for request/response containing
    # auth-related fields when ATS is configured to bypass caching for those
    authIgnoredReplayFile = "replay/auth-ignored.replay.yaml"

    def __init__(self):
        self.setupOriginServer()
        self.setupTS()

    def setupOriginServer(self):
        self.server = Test.MakeVerifierServerProcess("auth-ignored-verifier-server", self.authIgnoredReplayFile)

    def setupTS(self):
        self.ts = Test.MakeATSProcess("ts-auth-ignored")
        self.ts.Disk.records_config.update(
            {
                "proxy.config.diags.debug.enabled": 1,
                "proxy.config.diags.debug.tags": "http",
                # Configure ATS to ignore the WWW-Authenticate header in
                # response(allow caching of such response)
                "proxy.config.http.cache.ignore_authentication": 1
            })
        self.ts.Disk.remap_config.AddLine(f"map / http://127.0.0.1:{self.server.Variables.http_port}/",)

    def runTraffic(self):
        tr = Test.AddTestRun(
            "Verify the proper caching behavior for request/response containing auth-related fields when ATS is configured to bypass caching for those"
        )
        tr.AddVerifierClientProcess(
            "auth-ignored-client", self.authIgnoredReplayFile, http_ports=[self.ts.Variables.port], other_args='--thread-limit 1')
        tr.Processes.Default.StartBefore(self.server)
        tr.Processes.Default.StartBefore(self.ts)
        tr.StillRunningAfter = self.server
        tr.StillRunningAfter = self.ts

    def run(self):
        self.runTraffic()


AuthIgnoredTest().run()



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Verify session ID properties.
'''

Test.SkipUnless(Condition.HasCurlFeature('http2'))

# Configure the server.
server = Test.MakeOriginServer("server")
request_header = {"headers": "GET / HTTP/1.1\r\nHost: example.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
response_header = {
    "headers": "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Length:0\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
server.addResponse("sessionfile.log", request_header, response_header)

# Configure ATS. Disable the cache to simplify the test.
ts = Test.MakeATSProcess("ts", enable_tls=True, enable_cache=False)

ts.addDefaultSSLFiles()

Test.PrepareTestPlugin(
    os.path.join(Test.Variables.AtsBuildGoldTestsDir, 'continuations', 'plugins', '.libs', 'session_id_verify.so'), ts)

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'session_id_verify',
        'proxy.config.cache.enable_read_while_writer': 0,
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir),
    })

ts.Disk.remap_config.AddLine('map / http://127.0.0.1:{0}'.format(server.Variables.Port))

ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

#
# Run some HTTP/1 traffic.
#
tr = Test.AddTestRun("Perform HTTP/1 transactions")
cmd = 'curl -v -H "host:example.com" http://127.0.0.1:{0}'.format(ts.Variables.port)
numberOfRequests = 100
# Create a bunch of curl commands to be executed in parallel. Default.Process
# is set in SpawnCommands.  On Fedora 28/29, it seems that curl will
# occasionally timeout after a couple seconds and return exitcode 2
# Examining the packet capture shows that Traffic Server dutifully sends the response
ps = tr.SpawnCommands(cmdstr=cmd, count=numberOfRequests, retcode=Any(0, 2))
tr.Processes.Default.Env = ts.Env
tr.Processes.Default.ReturnCode = Any(0, 2)
tr.Processes.Default.StartBefore(server, ready=When.PortOpen(server.Variables.Port))
tr.Processes.Default.StartBefore(Test.Processes.ts)
ts.StartAfter(*ps)
server.StartAfter(*ps)

tr.StillRunningAfter = ts
tr.StillRunningAfter = server

#
# Run some HTTP/2 traffic.
#
tr = Test.AddTestRun("Perform HTTP/2 transactions")
cmd = 'curl -v -k --http2 -H "host:example.com" https://127.0.0.1:{0}'.format(ts.Variables.ssl_port)
ps = tr.SpawnCommands(cmdstr=cmd, count=numberOfRequests, retcode=Any(0, 2))
tr.Processes.Default.Env = ts.Env
tr.Processes.Default.ReturnCode = Any(0, 2)

#
# Verify that the session ids are unique.
#

# AuTest already searches for errors in diags.log and fails if it encounters
# them. The test plugin prints an error to this log if it sees duplicate ids.
# The following is to verify that we encountered the expected ids.


def verify_session_count(output):
    global numberOfRequests
    nReq = numberOfRequests * 2
    session_ids = [line[0:line.find("\n")] for line in str(output).split("session id: ")[1:]]
    if len(session_ids) != nReq:
        return "Found {} session_id's, expected {}".format(len(session_ids), nReq)
    return ""


ts.Disk.traffic_out.Content += Testers.FileContentCallback(verify_session_count, 'verify_session_count')



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = 'Test PROXY Protocol'
Test.ContinueOnFail = True


class ProxyProtocolInTest:
    """Test that ATS can receive Proxy Protocol."""

    replay_file = "replay/proxy_protocol_in.replay.yaml"

    def __init__(self):
        self.setupOriginServer()
        self.setupTS()

    def setupOriginServer(self):
        self.server = Test.MakeVerifierServerProcess("pp-in-server", self.replay_file)

    def setupTS(self):
        self.ts = Test.MakeATSProcess("ts_in", enable_tls=True, enable_cache=False, enable_proxy_protocol=True)

        self.ts.addDefaultSSLFiles()
        self.ts.Disk.ssl_multicert_config.AddLine("dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key")

        self.ts.Disk.remap_config.AddLine(f"map / http://127.0.0.1:{self.server.Variables.http_port}/")

        self.ts.Disk.records_config.update(
            {
                "proxy.config.http.proxy_protocol_allowlist": "127.0.0.1",
                "proxy.config.http.insert_forwarded": "for|by=ip|proto",
                "proxy.config.ssl.server.cert.path": f"{self.ts.Variables.SSLDir}",
                "proxy.config.ssl.server.private_key.path": f"{self.ts.Variables.SSLDir}",
                "proxy.config.diags.debug.enabled": 1,
                "proxy.config.diags.debug.tags": "proxyprotocol",
            })

        self.ts.Disk.logging_yaml.AddLines(
            '''
logging:
  formats:
    - name: access
      format: '%<chi> %<pps>'

  logs:
    - filename: access
      format: access
'''.split("\n"))

    def runTraffic(self):
        tr = Test.AddTestRun("Verify correct handling of incoming PROXY header.")
        tr.AddVerifierClientProcess(
            "pp-in-client",
            self.replay_file,
            http_ports=[self.ts.Variables.proxy_protocol_port],
            https_ports=[self.ts.Variables.proxy_protocol_ssl_port],
            other_args='--thread-limit 1')
        tr.Processes.Default.StartBefore(self.server)
        tr.Processes.Default.StartBefore(self.ts)
        tr.StillRunningAfter = self.server
        tr.StillRunningAfter = self.ts

    def checkAccessLog(self):
        """
        check access log
        """
        Test.Disk.File(os.path.join(self.ts.Variables.LOGDIR, 'access.log'), exists=True, content='gold/access.gold')

        # Wait for log file to appear, then wait one extra second to make sure
        # TS is done writing it.
        tr = Test.AddTestRun()
        tr.Processes.Default.Command = (
            os.path.join(Test.Variables.AtsTestToolsDir, 'condwait') + ' 60 1 -f ' +
            os.path.join(self.ts.Variables.LOGDIR, 'access.log'))
        tr.Processes.Default.ReturnCode = 0

    def run(self):
        self.runTraffic()
        self.checkAccessLog()


class ProxyProtocolOutTest:
    """Test that ATS can send Proxy Protocol."""

    _dns_counter = 0
    _client_counter = 0
    _server_counter = 0
    _ts_counter = 0
    _pp_out_replay_file = "replay/proxy_protocol_out.replay.yaml"

    def __init__(self, pp_version: int, is_tunnel: bool, is_tls_to_origin: bool = False) -> None:
        """Initialize a ProxyProtocolOutTest.

        :param pp_version: The Proxy Protocol version to use (1 or 2).
        :param is_tunnel: Whether ATS should tunnel to the origin.
        :param is_tls: Whether ATS should connect to the origin via TLS.
        """

        if pp_version not in (-1, 1, 2):
            raise ValueError(f'Invalid Proxy Protocol version (not 1 or 2): {pp_version}')
        self._pp_version = pp_version
        self._is_tunnel = is_tunnel
        self._is_tls_to_origin = is_tls_to_origin

    def setupOriginServer(self) -> None:
        """Configure the origin server.
        """
        self._server = Test.MakeVerifierServerProcess(
            f"pp-out-server-{ProxyProtocolOutTest._server_counter}", self._pp_out_replay_file)
        ProxyProtocolOutTest._server_counter += 1

    def setupDNS(self, tr: 'TestRun') -> None:
        """Configure the DNS server.

        :param tr: The TestRun to associate the DNS's Process with.
        """
        self._dns = tr.MakeDNServer(f'dns-{ProxyProtocolOutTest._dns_counter}', default='127.0.0.1')
        ProxyProtocolOutTest._dns_counter += 1

    def setupTS(self, tr: 'TestRun') -> None:
        """Configure Traffic Server."""
        process_name = f'ts-out-{ProxyProtocolOutTest._ts_counter}'
        ProxyProtocolOutTest._ts_counter += 1
        self._ts = tr.MakeATSProcess(process_name, enable_tls=True, enable_cache=False)

        self._ts.addDefaultSSLFiles()
        self._ts.Disk.ssl_multicert_config.AddLine("dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key")
        scheme = 'https' if self._is_tls_to_origin else 'http'
        server_port = self._server.Variables.https_port if self._is_tls_to_origin else self._server.Variables.http_port
        self._ts.Disk.remap_config.AddLine(f"map / {scheme}://backend.pp.origin.com:{server_port}/")

        self._ts.Disk.records_config.update(
            {
                "proxy.config.ssl.server.cert.path": f"{self._ts.Variables.SSLDir}",
                "proxy.config.ssl.server.private_key.path": f"{self._ts.Variables.SSLDir}",
                "proxy.config.diags.debug.enabled": 1,
                "proxy.config.diags.debug.tags": "http|proxyprotocol",
                "proxy.config.http.proxy_protocol_out": self._pp_version,
                "proxy.config.dns.nameservers": f"127.0.0.1:{self._dns.Variables.Port}",
                "proxy.config.dns.resolv_conf": 'NULL',
                "proxy.config.ssl.client.verify.server.policy": 'PERMISSIVE'
            })

        if self._is_tunnel:
            self._ts.Disk.records_config.update({
                "proxy.config.http.connect_ports": f'{self._server.Variables.https_port}',
            })

            self._ts.Disk.sni_yaml.AddLines(
                [
                    'sni:',
                    '- fqdn: pp.origin.com',
                    f'  tunnel_route: backend.pp.origin.com:{self._server.Variables.https_port}',
                ])

    def setLogExpectations(self, tr: 'TestRun') -> None:

        tr.Processes.Default.Streams.All += Testers.ContainsExpression("HTTP/1.1 200 OK", "Verify the client got a 200 response.")

        if self._pp_version in (1, 2):
            expected_pp = ('PROXY TCP4 127.0.0.1 127.0.0.1 '
                           rf'\d+ {self._ts.Variables.ssl_port}')
            self._server.Streams.All += Testers.ContainsExpression(
                expected_pp, "Verify the server got the expected Proxy Protocol string.")

            self._server.Streams.All += Testers.ContainsExpression(
                f'Received PROXY header v{self._pp_version}', "Verify the server got the expected Proxy Protocol version.")

        if self._pp_version == -1:
            self._server.Streams.All += Testers.ContainsExpression(
                'No valid PROXY header found', 'There should be no Proxy Protocol string.')

        if self._is_tunnel:
            self._ts.Disk.traffic_out.Content += Testers.ContainsExpression(
                'CONNECT tunnel://backend.pp.origin.com', 'Verify ATS establishes a blind tunnel to the server.')

    def run(self) -> None:
        """Run the test."""
        description = f'Proxy Protocol v{self._pp_version} '
        if self._is_tunnel:
            description += "with blind tunneling"
        else:
            description += "without blind tunneling"
            if self._is_tls_to_origin:
                description += " on TLS connection to origin"
        tr = Test.AddTestRun(description)

        self.setupDNS(tr)
        self.setupOriginServer()
        self.setupTS(tr)

        tr.AddVerifierClientProcess(
            f"pp-out-client-{ProxyProtocolOutTest._client_counter}",
            self._pp_out_replay_file,
            http_ports=[self._ts.Variables.port],
            https_ports=[self._ts.Variables.ssl_port],
            other_args='--thread-limit 1')
        ProxyProtocolOutTest._client_counter += 1
        self._ts.StartBefore(self._server)
        self._ts.StartBefore(self._dns)
        tr.Processes.Default.StartBefore(self._ts)
        tr.StillRunningAfter = self._server

        self.setLogExpectations(tr)


ProxyProtocolInTest().run()

# non-tunnling HTTP to origin
ProxyProtocolOutTest(pp_version=-1, is_tunnel=False, is_tls_to_origin=False).run()
ProxyProtocolOutTest(pp_version=1, is_tunnel=False, is_tls_to_origin=False).run()
ProxyProtocolOutTest(pp_version=2, is_tunnel=False, is_tls_to_origin=False).run()
# non-tunnling HTTPS to origin
ProxyProtocolOutTest(pp_version=-1, is_tunnel=False, is_tls_to_origin=True).run()
ProxyProtocolOutTest(pp_version=1, is_tunnel=False, is_tls_to_origin=True).run()
ProxyProtocolOutTest(pp_version=2, is_tunnel=False, is_tls_to_origin=True).run()
# tunneling
ProxyProtocolOutTest(pp_version=1, is_tunnel=True).run()
ProxyProtocolOutTest(pp_version=2, is_tunnel=True).run()



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Test ATS offering both RSA and EC certificates loaded via plugin
'''

Test.SkipUnless(Condition.HasOpenSSLVersion('1.1.1'))

# Define default ATS
ts = Test.MakeATSProcess("ts", enable_tls=True)
server = Test.MakeOriginServer("server", ssl=True)
dns = Test.MakeDNServer("dns")

request_header = {"headers": "GET / HTTP/1.1\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
response_header = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", request_header, response_header)

# add ssl materials like key, certificates for the server
ts.addSSLfile("ssl/signed-foo.pem")
ts.addSSLfile("ssl/signed2-foo.pem")
ts.addSSLfile("ssl/signed-foo.key")
ts.addSSLfile("ssl/signed-foo-ec.pem")
ts.addSSLfile("ssl/signed-foo-ec.key")
ts.addSSLfile("ssl/signed-san.pem")
ts.addSSLfile("ssl/signed-san.key")
ts.addSSLfile("ssl/signed-san-ec.pem")
ts.addSSLfile("ssl/signed-san-ec.key")
ts.addSSLfile("ssl/signer.pem")
ts.addSSLfile("ssl/signer.key")
ts.addSSLfile("ssl/server.pem")
ts.addSSLfile("ssl/server.key")

ts.Disk.remap_config.AddLine('map / https://foo.com:{1}'.format(ts.Variables.ssl_port, server.Variables.SSL_Port))

ts.Disk.ssl_multicert_config.AddLines(
    [
        'ssl_cert_name=signed-foo-ec.pem,signed-foo.pem ssl_key_name=signed-foo-ec.key,signed-foo.key',
        'ssl_cert_name=signed-san-ec.pem,signed-san.pem ssl_key_name=signed-san-ec.key,signed-san.key',
        'dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key',
    ])

Test.PrepareTestPlugin(os.path.join(Test.Variables.AtsTestPluginsDir, 'ssl_secret_load_test.so'), ts)

# Case 1, global config policy=permissive properties=signature
#         override for foo.com policy=enforced properties=all
ts.Disk.records_config.update(
    {
        'proxy.config.ssl.server.cert.path': '{0}/../'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}/../'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.cipher_suite':
            'ECDHE-RSA-AES128-GCM-SHA256:ECDHE-RSA-AES256-GCM-SHA384:ECDHE-RSA-AES128-SHA256:ECDHE-RSA-AES256-SHA384:AES128-GCM-SHA256:AES256-GCM-SHA384:ECDHE-RSA-RC4-SHA:ECDHE-RSA-AES128-SHA:ECDHE-RSA-AES256-SHA:RC4-SHA:RC4-MD5:AES128-SHA:AES256-SHA:DES-CBC3-SHA!SRP:!DSS:!PSK:!aNULL:!eNULL:!SSLv2',
        'proxy.config.url_remap.pristine_host_hdr': 1,
        'proxy.config.dns.nameservers': '127.0.0.1:{0}'.format(dns.Variables.Port),
        'proxy.config.exec_thread.autoconfig.scale': 1.0,
        'proxy.config.dns.resolv_conf': 'NULL',
        'proxy.config.diags.debug.tags': 'ssl_secret_load_test',
        'proxy.config.diags.debug.enabled': 1,
    })

dns.addRecords(records={"foo.com.": ["127.0.0.1"]})
dns.addRecords(records={"bar.com.": ["127.0.0.1"]})

# Should receive a EC cert
tr = Test.AddTestRun("Default for foo should return EC cert")
tr.Setup.Copy("ssl/signer.pem")
tr.Setup.Copy("ssl/signer2.pem")
tr.Processes.Default.Command = "echo foo | openssl s_client  -CAfile signer.pem -servername foo.com -connect 127.0.0.1:{0}".format(
    ts.Variables.ssl_port)
tr.ReturnCode = 0
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(dns)
tr.Processes.Default.StartBefore(Test.Processes.ts, ready=When.PortOpen(ts.Variables.ssl_port))
tr.StillRunningAfter = server
tr.StillRunningAfter = ts
tr.Processes.Default.Streams.All += Testers.ContainsExpression("Peer signature type: ECDSA", "Should select EC cert")
tr.Processes.Default.Streams.All += Testers.ExcludesExpression("unable to verify the first certificate", "Correct signer")

# Should receive a RSA cert
tr = Test.AddTestRun("Only offer RSA ciphers, should receive RSA cert")
tr.Processes.Default.Command = "echo foo | openssl s_client  -CAfile signer.pem -servername foo.com -sigalgs 'RSA-PSS+SHA256' -connect 127.0.0.1:{0}".format(
    ts.Variables.ssl_port)
tr.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts
tr.Processes.Default.Streams.All += Testers.ContainsExpression("Peer signature type: RSA-PSS", "Should select RSA cert")
tr.Processes.Default.Streams.All += Testers.ExcludesExpression("unable to verify the first certificate", "Correct signer")

# Should receive a EC cert
tr = Test.AddTestRun("Default for one.com should return EC cert")
tr.Processes.Default.Command = "echo foo | openssl s_client  -CAfile signer.pem -servername one.com -connect 127.0.0.1:{0}".format(
    ts.Variables.ssl_port)
tr.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts
tr.Processes.Default.Streams.All += Testers.ContainsExpression("Peer signature type: ECDSA", "Should select EC cert")
tr.Processes.Default.Streams.All += Testers.ContainsExpression("CN ?= ?group.com", "Should select a group SAN")
tr.Processes.Default.Streams.All += Testers.ExcludesExpression("unable to verify the first certificate", "Correct signer")

# Should receive a RSA cert
tr = Test.AddTestRun("Only offer RSA ciphers, should receive RSA cert")
tr.Processes.Default.Command = "echo foo | openssl s_client  -CAfile signer.pem -servername one.com -sigalgs 'RSA-PSS+SHA256' -connect 127.0.0.1:{0}".format(
    ts.Variables.ssl_port)
tr.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts
tr.Processes.Default.Streams.All += Testers.ContainsExpression("Peer signature type: RSA-PSS", "Should select RSA cert")
tr.Processes.Default.Streams.All += Testers.ContainsExpression("CN ?= ?group.com", "Should select a group SAN")
tr.Processes.Default.Streams.All += Testers.ExcludesExpression("unable to verify the first certificate", "Correct signer")

# Should receive a RSA cert
tr = Test.AddTestRun("rsa.com only in rsa cert")
tr.Processes.Default.Command = "echo foo | openssl s_client  -CAfile signer.pem -servername rsa.com -connect 127.0.0.1:{0}".format(
    ts.Variables.ssl_port)
tr.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts
tr.Processes.Default.Streams.All += Testers.ContainsExpression("Peer signature type: RSA-PSS", "Should select RSA cert")
tr.Processes.Default.Streams.All += Testers.ContainsExpression("CN ?= ?group.com", "Should select a group SAN")
tr.Processes.Default.Streams.All += Testers.ExcludesExpression("unable to verify the first certificate", "Correct signer")

# Should receive a EC cert
tr = Test.AddTestRun("ec.com only in ec cert")
tr.Processes.Default.Command = "echo foo | openssl s_client  -CAfile signer.pem -servername ec.com -connect 127.0.0.1:{0}".format(
    ts.Variables.ssl_port)
tr.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts
tr.Processes.Default.Streams.All += Testers.ContainsExpression("Peer signature type: ECDSA", "Should select EC cert")
tr.Processes.Default.Streams.All += Testers.ContainsExpression("CN ?= ?group.com", "Should select a group SAN")
tr.Processes.Default.Streams.All += Testers.ExcludesExpression("unable to verify the first certificate", "Correct signer")

# Copy in a new version of the foo.com cert.  Replace it with the version
# signed by signer 2.  Wait at least a second to sure the file update time
# differs
trupdate = Test.AddTestRun("Update server bar cert file in place")
trupdate.StillRunningAfter = ts
trupdate.StillRunningAfter = server
trupdate.Setup.CopyAs("ssl/signed2-foo.pem", ".", "{0}/signed-foo.pem".format(ts.Variables.SSLDir))
# For some reason the Setup.CopyAs does not change the modification time, so we touch
trupdate.Processes.Default.Command = 'touch {0}/signed-foo.pem'.format(ts.Variables.SSLDir)
# Need to copy over the environment so traffic_ctl knows where to find the unix domain socket
trupdate.Processes.Default.Env = ts.Env
trupdate.Processes.Default.ReturnCode = 0

# The plugin will pull every 3 seconds.  So wait 4 seconds and test again.  Request with CA=signer2.pem should work.  Request with CA=signer.pem should fail
# Should receive a RSA cert
tr = Test.AddTestRun("Only offer RSA ciphers, should receive RSA cert")
tr.Processes.Default.Command = "echo foo | openssl s_client -CAfile signer.pem  -servername foo.com -sigalgs 'RSA-PSS+SHA256' -connect 127.0.0.1:{0}".format(
    ts.Variables.ssl_port)
tr.DelayStart = 4
tr.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts
tr.Processes.Default.Streams.All += Testers.ContainsExpression("Peer signature type: RSA-PSS", "Should select RSA cert")
tr.Processes.Default.Streams.All += Testers.ContainsExpression("CN ?= ?foo.com", "Should select foo.com")
tr.Processes.Default.Streams.All += Testers.ContainsExpression("unable to verify the first certificate", "Different signer")

tr = Test.AddTestRun("Only offer RSA ciphers, should receive RSA cert with correct CA")
tr.Processes.Default.Command = "echo foo | openssl s_client -CAfile signer2.pem  -servername foo.com -sigalgs 'RSA-PSS+SHA256' -connect 127.0.0.1:{0}".format(
    ts.Variables.ssl_port)
tr.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts
tr.Processes.Default.Streams.All += Testers.ContainsExpression("Peer signature type: RSA-PSS", "Should select RSA cert")
tr.Processes.Default.Streams.All += Testers.ContainsExpression("CN ?= ?foo.com", "Should select foo.com")
tr.Processes.Default.Streams.All += Testers.ExcludesExpression("unable to verify the first certificate", "Correct signer")

# The EC case should be unchanged
tr = Test.AddTestRun("Offer any cipher")
tr.Processes.Default.Command = "echo foo | openssl s_client -CAfile signer.pem  -servername foo.com  -connect 127.0.0.1:{0}".format(
    ts.Variables.ssl_port)
tr.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts
tr.Processes.Default.Streams.All += Testers.ContainsExpression("Peer signature type: ECDSA", "Should select EC cert")
tr.Processes.Default.Streams.All += Testers.ContainsExpression("CN ?= ?foo.com", "Should select foo.com")
tr.Processes.Default.Streams.All += Testers.ExcludesExpression("unable to verify the first certificate", "Correct signer")



'''
Verify log file naming behavior.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import ports

Test.Summary = '''
Verify log file naming behavior.
'''


class LogFilenamesTest:
    """ Common test configuration logic across the filename tests.
    """

    # A counter for the ATS process to make each of them unique.
    __ts_counter = 1

    # The default log names for the various system logs.
    default_log_data = {'diags': 'diags.log', 'error': 'error.log'}

    def __init__(self, description, log_data=default_log_data):
        ''' Handle initialization tasks common across the tests.

        Args:
            description (str): The description of the test. This is passed to
            the TestRun.

            log_data (dict): The log name information passed to the
            MakeATSProcess extension.
        '''
        self.__description = description
        self.ts = self.__configure_traffic_server(log_data)
        self.tr = self.__configure_traffic_TestRun(description)
        self.__configure_await_TestRun(self.sentinel_log_path)

    def __configure_traffic_server(self, log_data):
        ''' Common ATS configuration logic.

        Args:
            log_data (dict): The log name information passed to the
            MakeATSProcess extension.

        Return:
            The traffic_server process.
        '''
        self._ts_name = f"ts{LogFilenamesTest.__ts_counter}"
        LogFilenamesTest.__ts_counter += 1
        self.ts = Test.MakeATSProcess(self._ts_name, use_traffic_out=False, log_data=log_data)
        self.ts.Disk.records_config.update(
            {
                'proxy.config.diags.debug.enabled': 0,
                'proxy.config.diags.debug.tags': 'log',
                'proxy.config.log.periodic_tasks_interval': 1,
            })

        # Intentionally retrieve a port that is closed, that is no server is
        # listening on it. We will use this to attempt talking with a
        # non-existent server, which will result in an error log entry.
        ports.get_port(self.ts, 'closed_port')
        self.ts.Disk.remap_config.AddLines(
            [
                f'map /server/down http://127.0.0.1:{self.ts.Variables.closed_port}',
                'map / https://trafficserver.apache.org @action=deny',
            ])

        # The following log is configured so that we can wait upon it being
        # written so we know that ATS is done writing logs.
        self.sentinel_log_filename = "sentinel"
        self.ts.Disk.logging_yaml.AddLine(
            f'''
            logging:
              formats:
                - name: url_and_return_code
                  format: "%<pqu>: %<pssc>"
              logs:
                - filename: {self.sentinel_log_filename}
                  format: url_and_return_code
            ''')

        self.sentinel_log_path = os.path.join(self.ts.Variables.LOGDIR, f"{self.sentinel_log_filename}.log")

        return self.ts

    def __configure_await_TestRun(self, log_path):
        ''' Configure a TestRun that awaits upon the provided log_path to
        exist.

        Args:
            log_path (str): The log file upon which we will wait.
        '''
        description = self.__description
        tr = Test.AddTestRun(f'Awaiting log files to be written for: {description}')
        condwait_path = os.path.join(Test.Variables.AtsTestToolsDir, 'condwait')
        tr.Processes.Default.Command = f'{condwait_path} 60 1 -f {log_path}'
        tr.Processes.Default.ReturnCode = 0

    def __configure_traffic_TestRun(self, description):
        ''' Configure a TestRun to run the expected transactions.

        Args:
            description (str): The description to use for the TestRun.
        '''
        tr = Test.AddTestRun(f'Run traffic for: {description}')
        tr.Processes.Default.Command = (
            f'curl http://127.0.0.1:{self.ts.Variables.port}/some/path --verbose --next '
            f'http://127.0.0.1:{self.ts.Variables.port}/server/down --verbose')
        tr.Processes.Default.ReturnCode = 0
        tr.Processes.Default.StartBefore(self.ts)

    def configure_named_custom_log(self, custom_log_filename):
        """ Configure ATS to log to the custom log file via logging.yaml.

        Args:
            custom_log_filename (str): The name of the custom log file to
            configure.

        Return:
            The path to the configured custom log file.
        """
        self.custom_log_filename = custom_log_filename
        self.ts.Disk.logging_yaml.AddLine(
            f'''
                - filename: {custom_log_filename}
                  format: url_and_return_code
            ''')

        if custom_log_filename in ('stdout', 'stderr'):
            self.custom_log_path = custom_log_filename
            if custom_log_filename == 'stdout':
                self.ts.Disk.custom_log = self.ts.Streams.stdout
            else:
                self.ts.Disk.custom_log = self.ts.Streams.stderr
        else:
            self.custom_log_path = os.path.join(self.ts.Variables.LOGDIR, f"{custom_log_filename}.log")
            self.ts.Disk.File(self.custom_log_path, id="custom_log")
        return self.custom_log_path

    def set_log_expectations(self):
        ''' Configure sanity checks for each of the log types (diags, error,
        etc.) to verify they are emitting the expected content.
        '''

        diags_path = self.ts.Disk.diags_log.AbsPath
        self.ts.Disk.diags_log.Content += Testers.ContainsExpression(
            "Traffic Server is fully initialized", f"{diags_path} should contain traffic_server diag messages")

        error_log_path = self.ts.Disk.error_log.AbsPath
        self.ts.Disk.error_log.Content += Testers.ContainsExpression(
            "CONNECT: attempt fail", f"{error_log_path} should contain connection error messages")

        custom_log_path = self.ts.Disk.custom_log.AbsPath
        self.ts.Disk.custom_log.Content += Testers.ContainsExpression(
            "https://trafficserver.apache.org/some/path: 403", f"{custom_log_path} should contain the custom transaction logs")


class DefaultNamedTest(LogFilenamesTest):
    ''' Verify that if custom names are not configured, then the default
    'diags.log' and 'error.log' are written to.
    '''

    def __init__(self):
        super().__init__('default log filename configuration')

        self.configure_named_custom_log('my_custom_log')
        self.set_log_expectations()


class CustomNamedTest(LogFilenamesTest):
    ''' Verify that the user can assign custom filenames to diags.log, etc.
    '''

    def __init__(self):
        log_data = {'diags': 'my_diags.log', 'error': 'my_error.log'}
        super().__init__('specify log filename configuration', log_data)

        # Configure custom names for diags.log, etc.
        self.ts.Disk.records_config.update(
            {
                'proxy.config.diags.logfile.filename': 'my_diags.log',
                'proxy.config.error.logfile.filename': 'my_error.log',
            })

        self.configure_named_custom_log('my_custom_log')
        self.set_log_expectations()


class stdoutTest(LogFilenamesTest):
    ''' Verify that we can configure the logs to go to stdout.
    '''

    def __init__(self):

        log_data = {'diags': 'stdout', 'error': 'stdout'}
        super().__init__('specify logs to go to stdout', log_data)

        # Configure custom names for diags.log, etc.
        self.ts.Disk.records_config.update(
            {
                'proxy.config.diags.logfile.filename': 'stdout',
                'proxy.config.error.logfile.filename': 'stdout',
            })

        self.configure_named_custom_log('stdout')

        # The diags.log file will not be created since we are piping to stdout.
        # Therefore, simply wait upon the port being open.
        self.ts.Ready = When.PortOpen(self.ts.Variables.port)
        self.set_log_expectations()


class stderrTest(LogFilenamesTest):
    '''
    Verify that we can configure the logs to go to stderr.
    '''

    def __init__(self):

        log_data = {'diags': 'stderr', 'error': 'stderr'}
        super().__init__('specify logs to go to stderr', log_data)

        # Configure custom names for diags.log, etc.
        self.ts.Disk.records_config.update(
            {
                'proxy.config.diags.logfile.filename': 'stderr',
                'proxy.config.error.logfile.filename': 'stderr',
            })

        self.configure_named_custom_log('stderr')

        # The diags.log file will not be created since we are piping to stderr.
        # Therefore, simply wait upon the port being open.
        self.ts.Ready = When.PortOpen(self.ts.Variables.port)
        self.set_log_expectations()


#
# Run the tests.
#
DefaultNamedTest()
CustomNamedTest()
stdoutTest()

# The following stderr test can be run successfully by hand using the replay
# files from the sandbox. All the expected output goes to stderr. However, for
# some reason during the AuTest run, the stderr output stops emitting after the
# logging.yaml file is parsed. This is left here for now because it is valuable
# for use during development, but it is left commented out so that it doesn't
# produce the false failure in CI and developer test runs.
# stderrTest()



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import sys

Test.Summary = '''
Test custom log file format
'''
Test.SkipUnless(Condition.HasATSFeature('TS_HAS_PIPE_BUFFER_SIZE_CONFIG'))

ts_counter = 1


def get_ts(logging_config):
    """
    Create a Traffic Server process.
    """
    global ts_counter
    ts = Test.MakeATSProcess("ts{}".format(ts_counter))
    ts_counter += 1

    ts.Disk.records_config.update(
        {
            'proxy.config.diags.debug.enabled': 1,
            'proxy.config.diags.debug.tags': 'log-file',
            'proxy.config.log.max_secs_per_buffer': 1,
        })

    # Since we're only verifying logs and not traffic, we don't need an origin
    # server. The following will simply deny the requests and emit a log
    # message.
    ts.Disk.remap_config.AddLine('map / http://www.linkedin.com/ @action=deny')

    ts.Disk.logging_yaml.AddLines(logging_config)

    return ts


#
# Test 1: Default configured log pipe size.
#
tr = Test.AddTestRun()
pipe_name = "default_pipe_size.pipe"
ts = get_ts(
    '''
logging:
  formats:
    - name: custom
      format: "%<hii> %<hiih>"
  logs:
    - filename: '{}'
      mode: ascii_pipe
      format: custom
'''.format(pipe_name).split("\n"))

pipe_path = os.path.join(ts.Variables.LOGDIR, pipe_name)

ts.Disk.traffic_out.Content += Testers.ContainsExpression(
    "Created named pipe .*{}".format(pipe_name), "Verify that the named pipe was created")

ts.Disk.traffic_out.Content += Testers.ContainsExpression(
    "no readers for pipe .*{}".format(pipe_name), "Verify that no readers for the pipe was detected.")

ts.Disk.traffic_out.Content += Testers.ExcludesExpression(
    "New buffer size for pipe".format(pipe_name), "Verify that the default pipe size was used.")

curl = tr.Processes.Process("client_request", 'curl "http://127.0.0.1:{0}" --verbose'.format(ts.Variables.port))

reader_output = os.path.join(ts.Variables.LOGDIR, "reader_output")
pipe_reader = tr.Processes.Process("pipe_reader", 'cat {} | tee {}'.format(pipe_path, reader_output))

# Create an arbitrary process that just sleeps so that we can provide a wait
# condition upon the log being emitted. The test won't wait this entire sleep
# period, it will only poll until the FileContains Ready condition is
# fulfilled.
wait_for_log = tr.Processes.Process("wait_for_log", 'sleep 15')
wait_for_log.Ready = When.FileContains(reader_output, '127.0.0.1')

# This is an arbitrary Default process that will simply provide for
# ordering of the Processes.
tr.Processes.Default.Command = "echo 'Default place holder for process ordering.'"
tr.Processes.Default.Return = 0

# Process ordering.
tr.Processes.Default.StartBefore(wait_for_log)
wait_for_log.StartBefore(curl)
curl.StartBefore(pipe_reader)
pipe_reader.StartBefore(ts)

#
# Test 2: Change the log's buffer size.
#
tr = Test.AddTestRun()
pipe_name = "change_pipe_size.pipe"
# 64 KB is the default, so set the size larger than that to verify we can
# increase the size.
pipe_size = 75000
ts = get_ts(
    '''
logging:
  formats:
    - name: custom
      format: "%<hii> %<hiih>"
  logs:
    - filename: '{}'
      mode: ascii_pipe
      format: custom
      pipe_buffer_size: {}
      '''.format(pipe_name, pipe_size).split("\n"))

pipe_path = os.path.join(ts.Variables.LOGDIR, pipe_name)

ts.Disk.traffic_out.Content += Testers.ContainsExpression(
    "Created named pipe .*{}".format(pipe_name), "Verify that the named pipe was created")

ts.Disk.traffic_out.Content += Testers.ContainsExpression(
    "no readers for pipe .*{}".format(pipe_name), "Verify that no readers for the pipe was detected.")

ts.Disk.traffic_out.Content += Testers.ContainsExpression(
    "Previous buffer size for pipe .*{}".format(pipe_name), "Verify that the named pipe's size was adjusted")

# See fcntl:
#   "Attempts to set the pipe capacity below the page size
#    are silently rounded up to the page size."
#
# As a result of this, we cannot check that the pipe size is the exact size we
# requested, but it should be at least that big. We use the
# pipe_buffer_is_larger_than.py helper script to verify that the pipe grew in
# size.
ts.Disk.traffic_out.Content += Testers.ContainsExpression(
    "New buffer size for pipe.*{}".format(pipe_name), "Verify that the named pipe's size was adjusted")
buffer_verifier = "pipe_buffer_is_larger_than.py"
tr.Setup.Copy(buffer_verifier)
verify_buffer_size = tr.Processes.Process("verify_buffer_size", f"{sys.executable} {buffer_verifier} {pipe_path} {pipe_size}")
verify_buffer_size.Return = 0
verify_buffer_size.Streams.All += Testers.ContainsExpression("Success", "The buffer size verifier should report success.")

curl = tr.Processes.Process("client_request", 'curl "http://127.0.0.1:{0}" --verbose'.format(ts.Variables.port))

reader_output = os.path.join(ts.Variables.LOGDIR, "reader_output")
pipe_reader = tr.Processes.Process("pipe_reader", 'cat {} | tee {}'.format(pipe_path, reader_output))

# Create an arbitrary process that just sleeps so that we can provide a wait
# condition upon the log being emitted. The test won't wait this entire sleep
# period, it will only poll until the FileContains Ready condition is
# fulfilled.
wait_for_log = tr.Processes.Process("wait_for_log", 'sleep 15')
wait_for_log.Ready = When.FileContains(reader_output, '127.0.0.1')

# This is an arbitrary Default process that will simply provide for
# ordering of the Processes.
tr.Processes.Default.Command = "echo 'Default place holder for process ordering.'"
tr.Processes.Default.Return = 0

# Process ordering.
tr.Processes.Default.StartBefore(verify_buffer_size)
verify_buffer_size.StartBefore(wait_for_log)
wait_for_log.StartBefore(curl)
curl.StartBefore(pipe_reader)
pipe_reader.StartBefore(ts)



#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Test adding hooks
'''

Test.ContinueOnFail = True

server = Test.MakeOriginServer("server")

request_header = {"headers": "GET /argh HTTP/1.1\r\nHost: doesnotmatter\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
response_header = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", request_header, response_header)

ts = Test.MakeATSProcess("ts", enable_cache=False)

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.tags': 'test',
        'proxy.config.diags.show_location': 0,
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.url_remap.remap_required': 0,
    })

Test.PrepareTestPlugin(os.path.join(Test.Variables.AtsTestPluginsDir, 'hook_add_plugin.so'), ts)

ts.Disk.remap_config.AddLine("map http://one http://127.0.0.1:{0}".format(server.Variables.Port))

tr = Test.AddTestRun()
# Probe server port to check if ready.
tr.Processes.Default.StartBefore(server, ready=When.PortOpen(server.Variables.Port))
tr.Processes.Default.StartBefore(Test.Processes.ts)
#
tr.Processes.Default.Command = ('curl --verbose --ipv4 --header "Host: one" http://localhost:{0}/argh'.format(ts.Variables.port))
tr.Processes.Default.ReturnCode = 0

# Look at the debug output from the plugin
ts.Disk.traffic_out.Content = "hook_add.gold"



#!/usr/bin/env python3

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import argparse
import random
import time

# https://github.com/mpdavis/python-jose
from jose import jwt


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--config', help="Configuration File", required=True)
    parser.add_argument('-u', '--uri', help="URI to sign", required=True)

    # helpers
    parser.add_argument('--key_index', type=int, nargs=1)
    parser.add_argument('--token_lifetime', type=int, nargs=1)

    # override arguments -- claims
    parser.add_argument('--aud', nargs=1)
    parser.add_argument('--cdniets', type=int, nargs=1)
    parser.add_argument('--cdnistd', type=int, nargs=1)
    parser.add_argument('--cdnistt', type=int, nargs=1)
    parser.add_argument('--exp', type=int, nargs=1)
    parser.add_argument('--iss', nargs=1)

    # override arguments -- key
    parser.add_argument('--alg', nargs=1)
    parser.add_argument('--k', nargs=1)
    parser.add_argument('--kid', nargs=1)
    parser.add_argument('--kty', nargs=1)

    args = parser.parse_args()

    with open(args.config, 'r') as f:
        config = json.load(f)

    keys = config["keys"]

    # Select a key, either explicitly or randomly
    key_index = 0
    if args.key_index:
        key_index = args.key_index[0]
        print("args key_index " + str(key_index))
    else:
        key_index = random.randint(0, len(keys) - 1)
        print("randomizing key index")

    print("Using key_index " + str(key_index))

    print("Using Key: " + str(keys[key_index]["kid"]) + " to sign URI.")
    key = keys[key_index]

    # Build Out claimset
    claimset = {}
    if "iss" in config.keys():
        claimset["iss"] = config["iss"]

    if "token_lifetime" in config.keys():
        claimset["exp"] = int(time.time()) + config["token_lifetime"]
    else:
        claimset["exp"] = int(time.time()) + 30

    if "aud" in config.keys():
        claimset["aud"] = config["aud"]

    if "cdnistt" in config.keys():
        if config["cdnistt"]:
            claimset["cdnistt"] = 1
            if "cdniets" in config.keys():
                claimset["cdniets"] = config["cdniets"]
            else:
                claimset["cdniets"] = 30

    if "cdnistd" in config.keys():
        claimset["cdnistd"] = config["cdnistd"]

    # process override args - simple
    if args.iss:
        claimset["iss"] = args.iss[0]
    if args.exp:
        claimset["exp"] = args.exp[0]
    if args.aud:
        claimset["aud"] = args.aud[0]
    if args.cdnistd:
        claimset["cdnistd"] = args.cdnistd[0]

    # process override args - complex
    if args.cdnistt:
        claimset["cdnistt"] = args.cdnistt[0]

    if "cdnistt" in config.keys():
        if args.cdniets:
            claimset["cdniets"] = arg.cdniets[0]

    # specific key overrides
    if args.alg:
        key["alg"] = args.alg[0]
    if args.kid:
        key["kid"] = args.kid[0]
    if args.kty:
        key["kty"] = args.kty[0]
    if args.k:
        key["k"] = args.k[0]

    print(claimset)
    print(key)

    Token = jwt.encode(claimset, key, algorithm=key["alg"])

    print("Signed URL: " + args.uri + "?URISigningPackage=" + Token)


if __name__ == "__main__":
    main()



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Test lua functionality
'''

Test.SkipUnless(Condition.PluginExists('tslua.so'),)

Test.ContinueOnFail = True
# Define default ATS
ts = Test.MakeATSProcess("ts")
server = Test.MakeOriginServer("server")

Test.testName = ""
request_header = {"headers": "GET / HTTP/1.1\r\nHost: www.example.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
# expected response from the origin server
response_header = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": ""}

# add response to the server dictionary
server.addResponse("sessionfile.log", request_header, response_header)

ts.Disk.remap_config.AddLine('map / http://127.0.0.1:{}/'.format(server.Variables.Port) + ' @plugin=tslua.so @pparam=watermark.lua')

# Configure the tslua's configuration file.
ts.Setup.Copy("watermark.lua", ts.Variables.CONFIGDIR)

ts.Disk.records_config.update({'proxy.config.diags.debug.enabled': 1, 'proxy.config.diags.debug.tags': 'ts_lua'})

# Test for watermark debug output
ts.Disk.traffic_out.Content = Testers.ContainsExpression(r"WMbytes\(31337\)", "Upstream watermark should be properly set")

# Test if watermark upstream is set
tr = Test.AddTestRun("Lua Watermark")
tr.Processes.Default.Command = "curl -v http://127.0.0.1:{0}".format(ts.Variables.port)
tr.Processes.Default.StartBefore(server, ready=When.PortOpen(server.Variables.Port))

tr.Processes.Default.StartBefore(ts)
tr.StillRunningAfter = server



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from enum import Enum
import os
import re

Test.Summary = 'Exercise HTTP CONNECT Method'
Test.ContinueOnFail = True


class ConnectTest:

    class State(Enum):
        """
        State of process
        """
        INIT = 0
        RUNNING = 1

    def __init__(self):
        self.state = self.State.INIT
        self.__setupOriginServer()
        self.__setupTS()

    def __setupOriginServer(self):
        self.httpbin = Test.MakeHttpBinServer("httpbin")

    def __setupTS(self):
        self.ts = Test.MakeATSProcess("ts")

        self.ts.Disk.records_config.update(
            {
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'http',
                'proxy.config.http.server_ports': f"{self.ts.Variables.port}",
                'proxy.config.http.connect_ports': f"{self.httpbin.Variables.Port}",
            })

        self.ts.Disk.remap_config.AddLines([
            f"map http://foo.com/ http://127.0.0.1:{self.httpbin.Variables.Port}/",
        ])

        self.ts.Disk.logging_yaml.AddLines(
            '''
logging:
  formats:
    - name: common
      format: '%<chi> - %<caun> [%<cqtn>] "%<cqhm> %<pqu> %<cqpv>" %<pssc> %<pscl>'
  logs:
    - filename: access
      format: common
'''.split("\n"))

    def __checkProcessBefore(self, tr):
        if self.state == self.State.RUNNING:
            tr.StillRunningBefore = self.httpbin
            tr.StillRunningBefore = self.ts
        else:
            tr.Processes.Default.StartBefore(self.httpbin, ready=When.PortOpen(self.httpbin.Variables.Port))
            tr.Processes.Default.StartBefore(self.ts)
            self.state = self.State.RUNNING

    def __checkProcessAfter(self, tr):
        assert (self.state == self.State.RUNNING)
        tr.StillRunningAfter = self.httpbin
        tr.StillRunningAfter = self.ts

    def __testCase0(self):
        tr = Test.AddTestRun()
        self.__checkProcessBefore(tr)
        tr.Processes.Default.Command = f"curl -v --fail -s -p -x 127.0.0.1:{self.ts.Variables.port} 'http://foo.com/get'"
        tr.Processes.Default.ReturnCode = 0
        tr.Processes.Default.Streams.stderr = "gold/connect_0_stderr.gold"
        tr.Processes.Default.TimeOut = 3
        self.__checkProcessAfter(tr)

    def __testAccessLog(self):
        """Wait for log file to appear, then wait one extra second to make sure TS is done writing it."""
        Test.Disk.File(os.path.join(self.ts.Variables.LOGDIR, 'access.log'), exists=True, content='gold/connect_access.gold')

        tr = Test.AddTestRun()
        tr.Processes.Default.Command = (
            os.path.join(Test.Variables.AtsTestToolsDir, 'condwait') + ' 60 1 -f ' +
            os.path.join(self.ts.Variables.LOGDIR, 'access.log'))
        tr.Processes.Default.ReturnCode = 0

    def run(self):
        self.__testCase0()
        self.__testAccessLog()


ConnectTest().run()


class ConnectViaPVTest:
    # This test also executes the CONNECT request but using proxy verifier to
    # generate traffic
    connectReplayFile = "replays/connect.replay.yaml"

    def __init__(self):
        self.setupOriginServer()
        self.setupTS()

    def setupOriginServer(self):
        self.server = Test.MakeVerifierServerProcess("connect-verifier-server", self.connectReplayFile)
        # Verify server output
        self.server.Streams.stdout += Testers.ExcludesExpression("uuid: 1", "Verify the CONNECT request doesn't reach the server.")
        self.server.Streams.stdout += Testers.ContainsExpression(
            "GET /get HTTP/1.1\nuuid: 2", reflags=re.MULTILINE, description="Verify the server gets the second request.")

    def setupTS(self):
        self.ts = Test.MakeATSProcess("connect-ts")

        self.ts.Disk.records_config.update(
            {
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'http|iocore_net|rec',
                'proxy.config.http.server_ports': f"{self.ts.Variables.port}",
                'proxy.config.http.connect_ports': f"{self.server.Variables.http_port}",
            })

        self.ts.Disk.remap_config.AddLines([
            f"map / http://127.0.0.1:{self.server.Variables.http_port}/",
        ])
        # Verify ts logs
        self.ts.Disk.traffic_out.Content += Testers.ContainsExpression(
            f"Proxy's Request.*\n.*\nCONNECT 127.0.0.1:{self.server.Variables.http_port} HTTP/1.1",
            reflags=re.MULTILINE,
            description="Verify that ATS recognizes the CONNECT request.")

    def runTraffic(self):
        tr = Test.AddTestRun("Verify correct handling of CONNECT request")
        tr.AddVerifierClientProcess(
            "connect-client", self.connectReplayFile, http_ports=[self.ts.Variables.port], other_args='--thread-limit 1')
        tr.Processes.Default.StartBefore(self.server)
        tr.Processes.Default.StartBefore(self.ts)
        tr.StillRunningAfter = self.server
        tr.StillRunningAfter = self.ts

    def __testMetrics(self):
        tr = Test.AddTestRun("Test metrics")
        tr.Processes.Default.Command = (
            f"{Test.Variables.AtsTestToolsDir}/stdout_wait" + " 'traffic_ctl metric get" +
            " proxy.process.http.total_incoming_connections" + " proxy.process.http.total_client_connections" +
            " proxy.process.http.total_client_connections_ipv4" + " proxy.process.http.total_client_connections_ipv6" +
            " proxy.process.http.total_server_connections" + " proxy.process.http2.total_client_connections" +
            " proxy.process.http.connect_requests" + " proxy.process.tunnel.total_client_connections_blind_tcp" +
            " proxy.process.tunnel.current_client_connections_blind_tcp" +
            " proxy.process.tunnel.total_server_connections_blind_tcp" +
            " proxy.process.tunnel.current_server_connections_blind_tcp" +
            " proxy.process.tunnel.total_client_connections_tls_tunnel" +
            " proxy.process.tunnel.current_client_connections_tls_tunnel" +
            " proxy.process.tunnel.total_client_connections_tls_forward" +
            " proxy.process.tunnel.current_client_connections_tls_forward" +
            " proxy.process.tunnel.total_client_connections_tls_partial_blind" +
            " proxy.process.tunnel.current_client_connections_tls_partial_blind" +
            " proxy.process.tunnel.total_client_connections_tls_http" +
            " proxy.process.tunnel.current_client_connections_tls_http" + " proxy.process.tunnel.total_server_connections_tls" +
            " proxy.process.tunnel.current_server_connections_tls'" + f" {Test.TestDirectory}/gold/metrics.gold")
        # Need to copy over the environment so traffic_ctl knows where to find the unix domain socket
        tr.Processes.Default.Env = self.ts.Env
        tr.Processes.Default.ReturnCode = 0
        tr.StillRunningAfter = self.server
        tr.StillRunningAfter = self.ts

    def run(self):
        self.runTraffic()
        self.__testMetrics()


ConnectViaPVTest().run()


class ConnectViaPVTest2:
    # This test executes a HTTP/2 CONNECT request with Proxy Verifier.
    connectReplayFile = "replays/connect_h2.replay.yaml"

    def __init__(self):
        self.setupOriginServer()
        self.setupTS()

    def setupOriginServer(self):
        self.server = Test.MakeVerifierServerProcess("connect-verifier-server2", self.connectReplayFile)
        # Verify server output
        self.server.Streams.stdout += Testers.ExcludesExpression(
            "test: connect-request", "Verify the CONNECT request doesn't reach the server.")
        self.server.Streams.stdout += Testers.ContainsExpression(
            "GET /get HTTP/1.1\nuuid: 1\ntest: real-request",
            reflags=re.MULTILINE,
            description="Verify the server gets the second(tunneled) request.")

    def setupTS(self):
        self.ts = Test.MakeATSProcess("connect-ts2", enable_tls=True)

        self.ts.Disk.records_config.update(
            {
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'http|hpack',
                'proxy.config.ssl.server.cert.path': f'{self.ts.Variables.SSLDir}',
                'proxy.config.ssl.server.private_key.path': f'{self.ts.Variables.SSLDir}',
                'proxy.config.http.server_ports': f"{self.ts.Variables.ssl_port}:ssl",
                'proxy.config.http.connect_ports': f"{self.server.Variables.http_port}",
            })

        self.ts.addDefaultSSLFiles()
        self.ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

        self.ts.Disk.remap_config.AddLines([
            f"map / http://127.0.0.1:{self.server.Variables.http_port}/",
        ])
        # Verify ts logs
        self.ts.Disk.traffic_out.Content += Testers.ContainsExpression(
            f"Proxy's Request.*\n.*\nCONNECT 127.0.0.1:{self.server.Variables.http_port} HTTP/1.1",
            reflags=re.MULTILINE,
            description="Verify that ATS recognizes the CONNECT request.")

    def runTraffic(self):
        tr = Test.AddTestRun("Verify correct handling of CONNECT request on HTTP/2")
        tr.AddVerifierClientProcess(
            "connect-client2", self.connectReplayFile, https_ports=[self.ts.Variables.ssl_port], other_args='--thread-limit 1')
        tr.Processes.Default.StartBefore(self.server)
        tr.Processes.Default.StartBefore(self.ts)
        tr.StillRunningAfter = self.server
        tr.StillRunningAfter = self.ts

    def run(self):
        self.runTraffic()


ConnectViaPVTest2().run()



#!/usr/bin/env python3
'''
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
'''

from collections import defaultdict
from copy import deepcopy
from queue import Queue
from threading import Thread
import argparse
import json
import logging
import os
import sys

description = '''
Post process replay files produced by traffic_dump and clean it of any
incomplete transactions or sessions which got partially written because Traffic
Server was interrupted mid-connection. This also merges sessions to the same
client and, by default, formats the output files with human readable spacing.
'''

# Base replay file template with basic elements
TEMPLATE = json.loads('{"meta": {"version":"1.0"},"sessions":[]}')


class PostProcessError(Exception):
    ''' Base class for post processing errors.
    '''

    def __init__(self, message=None):
        self.message = message

    def __str__(self, *args):
        if self.message:
            return self.message
        else:
            return 'PostProcessError raised'


class VerifyError(PostProcessError):
    ''' Base class for node node verification errors.
    '''
    pass


class VerifyRequestError(VerifyError):
    ''' There was a problem verifying a request node.
    '''
    pass


class VerifyResponseError(VerifyError):
    ''' There was a problem verifying a response node.
    '''
    pass


class VerifySessionError(VerifyError):
    ''' There was a problem verifying a session node.
    '''
    pass


def verify_request(request):
    """ Function to verify request with method, url, and headers
    Args:
        request (json object)

    Raises:
        VerifyRequestError if there is a problem with the request.
    """
    if not request:
        raise VerifyRequestError('No request found.')
    if "method" not in request or not request["method"]:
        raise VerifyRequestError("Request did not have a method.")
    if "url" not in request or not request["url"]:
        raise VerifyRequestError("Request did not have a url.")
    if "headers" not in request or not request["headers"]:
        raise VerifyRequestError("Request did not have headers.")


def verify_response(response):
    """ Function to verify response with status
    Args:
        response (json object)

    Raises:
        VerifyResponseError if there is a problem with the response.
    """
    if not response:
        raise VerifyResponseError("No response found.")
    if "status" not in response or not response["status"]:
        raise VerifyResponseError("Response did not have a status.")


def verify_transaction(transaction, fabricate_proxy_requests=False):
    """ Function to verify that a transaction looks complete.

    Args:
        transaction (json object)
        fabricate_proxy_requests (bool) Whether the post-processor should
          fabricate proxy requests if they don't exist because the proxy served
          the response locally.

    Raises:
        VerifySessionError if there is no transaction.
        VerifyRequestError if there is a problem with a request.
        VerifyResponseError if there is a problem with a response.
    """
    if not transaction:
        raise VerifySessionError('No transaction found in the session.')

    if "client-request" not in transaction:
        raise VerifyRequestError('client-request not found in transaction')
    else:
        verify_request(transaction["client-request"])

    if "proxy-request" not in transaction and fabricate_proxy_requests:
        if "proxy-response" not in transaction:
            raise VerifyRequestError('proxy-response not found in transaction with a client-request')
        transaction["proxy-request"] = transaction["client-request"]
        if "server-response" not in transaction:
            transaction["server-response"] = transaction["proxy-response"]

    # proxy-response nodes can be empty.
    if "proxy-response" not in transaction:
        raise VerifyResponseError('proxy-response not found in transaction')

    if "proxy-request" in transaction or "server-response" in transaction:
        # proxy-request nodes can be empty, so no need to verify_response.
        if "proxy-request" not in transaction:
            raise VerifyRequestError('proxy-request not found in transaction')

        if "server-response" not in transaction:
            raise VerifyResponseError('server-response not found in transaction')
        else:
            verify_response(transaction["server-response"])


def verify_session(session, fabricate_proxy_requests=False):
    """ Function to verify that a session looks complete.

        A valid session contains a valid list of transactions.

    Args:
        transaction (json object)
        fabricate_proxy_requests (bool) Whether the post-processor should
          fabricate proxy requests if they don't exist because the proxy served
          the response locally.

    Raises:
        VerifyError if there is a problem with the session.
    """
    if not session:
        raise VerifySessionError('Session not found.')
    if "transactions" not in session or not session["transactions"]:
        raise VerifySessionError('No transactions found in session.')
    for transaction in session["transactions"]:
        verify_transaction(transaction, fabricate_proxy_requests)


def write_sessions(sessions, filename, indent):
    """ Write the JSON sessions to the given filename.

    Args:
        sessions The parsed JSON sessions to dump into filename.
        filename (string) The path to the file to write the parsed JSON file to.
        indent (int) The number of spaces per line to write to the file. A
          value of None causes the whole JSON file to be written as a single line.
    """
    new_json = deepcopy(TEMPLATE)
    new_json["sessions"] = deepcopy(sessions)
    with open(filename, "w") as f:
        json.dump(new_json, f, ensure_ascii=False, indent=indent)
        logging.debug(f"{filename} has {len(sessions)} sessions")


class ParseJSONError(PostProcessError):
    ''' There was an error opening or parsing the replay file.
    '''
    pass


def parse_json(replay_file):
    """ Open and parse the replay_file.

    Args:
        replay_file (string) The file with JSON content to parse.

    Return:
        The json package parsed JSON file or None if there was a problem
        parsing the file.
    """
    try:
        fd = open(replay_file, 'r')
    except Exception as e:
        logging.exception("Failed to open %s.", replay_file)
        raise ParseJSONError(e)

    try:
        parsed_json = json.load(fd)
    except Exception as e:
        message = e.msg.split(':')[0]
        logging.error("Failed to load %s as a JSON object: %s", replay_file, e)
        raise ParseJSONError(message)

    return parsed_json


def readAndCombine(replay_dir, num_sessions_per_file, indent, fabricate_proxy_requests, out_dir):
    """ Read raw dump files, filter out incomplete sessions, and merge
    them into output files.

    Args:
        replay_dir (string) Full path to dumps
        num_sessions_per_file (int) number of sessions in each output file
        indent (int) The number of spaces per line in the output replay files.
        fabricate_proxy_requests (bool) Whether the post-processor should
          fabricate proxy requests if they don't exist because the proxy served
          the response locally.
        out_dir (string) Output directory for post-processed json files.
    """
    session_count = 0
    batch_count = 0
    transaction_count = 0
    error_count = defaultdict(int)

    base_name = os.path.basename(replay_dir)

    sessions = []
    for f in os.listdir(replay_dir):
        replay_file = os.path.join(replay_dir, f)
        if not os.path.isfile(replay_file):
            continue

        try:
            parsed_json = parse_json(replay_file)
        except ParseJSONError as e:
            error_count[e.message] += 1
            continue

        for session in parsed_json["sessions"]:
            try:
                verify_session(session, fabricate_proxy_requests)
            except VerifyError as e:
                connection_time = session['connection-time']
                if not connection_time:
                    connection_time = session['start-time']
                if connection_time:
                    logging.debug("Omitting session in %s with connection-time: %d: %s", replay_file, session['connection-time'], e)
                else:
                    logging.debug("Omitting a session in %s, could not find a connection time: %s", replay_file, e)
                continue
            sessions.append(session)
            session_count += 1
            transaction_count += len(session["transactions"])
        if len(sessions) >= num_sessions_per_file:
            write_sessions(sessions, f"{out_dir}/{base_name}_{batch_count}.json", indent)
            sessions = []
            batch_count += 1
    if sessions:
        write_sessions(sessions, f"{out_dir}/{base_name}_{batch_count}.json", indent)

    return session_count, transaction_count, error_count


def post_process(in_dir, subdir_q, out_dir, num_sessions_per_file, single_line, fabricate_proxy_requests, cnt_q):
    """ Function used to set up individual threads.

    Each thread loops over the subdir_q, pulls a directory from there, and
    process the replay files in that directory. The threads finish when the
    subdir queue is empty, meaning each subdir has been processed.

    Args:
        in_dir (string) Path to parent of the subdirectories in subdir_q.
        subdir_q (Queue) Queue of subdir to read from.
        out_dir (string) The directory into which the post processed replay files
          are placed.
        num_sessions_per_file (int) traffic_dump will emit a separate file per
          session. This mechanism merges sessions within a single subdir.
          num_sessions_per_file is the limit to the number of sessions merged
          into a single replay file.
        single_line (bool) Whether to emit replay files as a single line. If
          false, the file is spaced out in a human readable fashion.
        fabricate_proxy_requests (bool) Whether the post-processor should
          fabricate proxy requests if they don't exist because the proxy served
          the response locally.
        cnt_q (Queue) Session, transaction, error count queue populated by each
          thread.
    """
    while not subdir_q.empty():
        subdir = subdir_q.get()
        subdir_path = os.path.join(in_dir, subdir)
        indent = 2
        if single_line:
            indent = None
        cnt = readAndCombine(subdir_path, num_sessions_per_file, indent, fabricate_proxy_requests, out_dir)
        cnt_q.put(cnt)


def configure_logging(use_debug=False):
    ''' Configure the logging mechanism.

    Args:
        use_debug (bool) Whether to configure debug-level logging.
    '''
    log_format = '%(levelname)s: %(message)s'
    if use_debug:
        logging.basicConfig(format=log_format, level=logging.DEBUG)
    else:
        logging.basicConfig(format=log_format, level=logging.INFO)


def parse_args():
    ''' Parse the command line arguments.
    '''
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument(
        "in_dir",
        type=str,
        help='''The input directory of traffic_dump replay
                        files.  The expectation is that this will contain
                        sub-directories that themselves contain replay files.
                        This is written to accommodate the directory populated
                        by traffic_dump via the --logdir option.''')
    parser.add_argument("out_dir", type=str, help="The output directory of post processed replay files.")
    parser.add_argument(
        "-n",
        "--num_sessions",
        type=int,
        default=10,
        help='''The maximum number of sessions merged into
                        single replay output files. The default is 10.''')
    parser.add_argument(
        "--no-human-readable",
        action="store_true",
        help='''By default, post processor will generate replay
                        files that are spaced out in a human readable format.
                        This turns off that behavior and leaves the files as
                        single-line entries.''')
    parser.add_argument(
        "--no-fabricate-proxy-requests",
        action="store_true",
        help='''By default, post processor will fabricate proxy
                        requests and server responses for transactions served
                        out of the proxy. Presumably in replay conditions,
                        these fabricated requests and responses will not hurt
                        anything because the Proxy Verifier server will not
                        notice if the proxy replies locally in replay
                        conditions. However, if it doesn't reply locally, then
                        the server will not know how to reply to these
                        requests. Using this option turns off this fabrication
                        behavior.''')
    parser.add_argument("-j", "--num_threads", type=int, default=32, help='''The maximum number of threads to use.''')
    parser.add_argument("-d", "--debug", action="store_true", help="Enable debug level logging.")
    return parser.parse_args()


def main():
    args = parse_args()
    configure_logging(use_debug=args.debug)
    logging.debug("Original options: %s", " ".join(sys.argv))

    if not os.path.exists(args.out_dir):
        os.mkdir(args.out_dir)

    # generate thread arguments
    subdir_q = Queue()
    cnt_q = Queue()
    for subdir in os.listdir(args.in_dir):
        if os.path.isdir(os.path.join(args.in_dir, subdir)):
            subdir_q.put(subdir)

    threads = []
    nthreads = min(max(subdir_q.qsize(), 1), args.num_threads)

    # Start up the threads.
    for _ in range(nthreads):
        t = Thread(
            target=post_process,
            args=(
                args.in_dir, subdir_q, args.out_dir, args.num_sessions, args.no_human_readable,
                not args.no_fabricate_proxy_requests, cnt_q))
        t.start()
        threads.append(t)

    # Wait for them to finish.
    for t in threads:
        t.join()

    # Retrieve the counts
    session_count = 0
    transaction_count = 0
    errors = defaultdict(int)
    for count_tuple in list(cnt_q.queue):
        session_count += count_tuple[0]
        transaction_count += count_tuple[1]
        for e in count_tuple[2]:
            errors[e] += count_tuple[2][e]
    summary = f"Total {session_count} sessions and {transaction_count} transactions."
    logging.info(summary)
    if errors:
        logging.info("Total errors:")
        for e in errors:
            logging.info(f"{e}: {errors[e]}")
    else:
        logging.info("Total errors: 0")

    with open(f"{args.out_dir}/summary.txt", "w", encoding="ascii") as f:
        f.write(f"{summary}\n")


if __name__ == "__main__":
    sys.exit(main())



#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Test reloading sni.yaml behaves as expected
'''

sni_domain = 'example.com'

ts = Test.MakeATSProcess("ts", enable_tls=True)
server = Test.MakeOriginServer("server")
server2 = Test.MakeOriginServer("server3")
request_header = {"headers": f"GET / HTTP/1.1\r\nHost: {sni_domain}\r\n\r\n", "timestamp": "1469733493.993", "body": ""}

response_header = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", request_header, response_header)

ts.Disk.records_config.update(
    {
        'proxy.config.ssl.server.cert.path': ts.Variables.SSLDir,
        'proxy.config.ssl.server.private_key.path': ts.Variables.SSLDir,
        'proxy.config.ssl.CA.cert.filename': f'{ts.Variables.SSLDir}/signer.pem',
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'ssl|http',
        'proxy.config.diags.output.debug': 'L',
    })

ts.addDefaultSSLFiles()
ts.addSSLfile("ssl/signed-foo.pem")
ts.addSSLfile("ssl/signed-foo.key")
ts.addSSLfile("ssl/signer.pem")

ts.Disk.remap_config.AddLine(f'map / http://127.0.0.1:{server.Variables.Port}')

ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

ts.Disk.sni_yaml.AddLines(
    f"""
      sni:
      - fqdn: {sni_domain}
        http2: off
        client_cert: {ts.Variables.SSLDir}/signed-foo.pem
        client_key: {ts.Variables.SSLDir}/signed-foo.key
        verify_client: STRICT
      """.split('\n'))

tr = Test.AddTestRun(f'ensure we can connect for SNI {sni_domain}')
tr.Setup.Copy("ssl/signed-foo.pem")
tr.Setup.Copy("ssl/signed-foo.key")
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.Processes.Default.StartBefore(server)
tr.StillRunningAfter = ts
tr.StillRunningAfter = server
tr.Processes.Default.Command = f"curl -q --tls-max 1.2 -s -v -k  --cert ./signed-foo.pem --key ./signed-foo.key --resolve '{sni_domain}:{ts.Variables.ssl_port}:127.0.0.1' https://{sni_domain}:{ts.Variables.ssl_port}"

tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Streams.stdout = Testers.ExcludesExpression("Could Not Connect", "Verify curl could successfully connect")
tr.Processes.Default.Streams.stderr = Testers.IncludesExpression(f"CN={sni_domain}", f"Verify curl used the {sni_domain} SNI")
ts.Disk.diags_log.Content = Testers.IncludesExpression(
    "SSL negotiation finished successfully", "Verify that the TLS handshake was successful")

# This config reload should fail because it references non-existent TLS key files
trupd = Test.AddTestRun("Update config file")
# Update the configs - this will overwrite the sni.yaml file
sniyamlpath = ts.Disk.sni_yaml.AbsPath
trupd.Disk.File(sniyamlpath, id="sni_yaml", typename="ats:config")
trupd.Disk.sni_yaml.AddLines(
    f"""
      sni:
      - fqdn: {sni_domain}
        http2: on
        client_cert: {ts.Variables.SSLDir}/signed-notexist.pem
        client_key: {ts.Variables.SSLDir}/signed-notexist.key
        verify_client: STRICT
      """.split('\n'))

trupd.StillRunningAfter = ts
trupd.StillRunningAfter = server
trupd.Processes.Default.Command = 'echo Updated configs'
trupd.Processes.Default.Env = ts.Env
trupd.Processes.Default.ReturnCode = 0

tr2reload = Test.AddTestRun("Reload config")
tr2reload.StillRunningAfter = ts
tr2reload.StillRunningAfter = server
tr2reload.Processes.Default.Command = 'traffic_ctl config reload'
tr2reload.Processes.Default.Env = ts.Env
tr2reload.Processes.Default.ReturnCode = 0
ts.Disk.diags_log.Content = Testers.ContainsExpression(
    'sni.yaml failed to load', 'reload should result in failure to load sni.yaml')

tr3 = Test.AddTestRun(f"Make request again for {sni_domain} that should still work")
# Wait for the reload to complete
tr3.Setup.Copy("ssl/signed-bar.pem")
tr3.Setup.Copy("ssl/signed-bar.key")
tr3.Processes.Default.StartBefore(server2, ready=When.FileContains(ts.Disk.diags_log.Name, "signed-notexist.pem", 1))
tr3.StillRunningAfter = ts
tr3.StillRunningAfter = server
tr3.Processes.Default.Command = f"curl -q --tls-max 1.2 -s -v -k  --cert ./signed-bar.pem --key ./signed-bar.key --resolve '{sni_domain}:{ts.Variables.ssl_port}:127.0.0.1' https://{sni_domain}:{ts.Variables.ssl_port}"

tr3.Processes.Default.ReturnCode = 0
# since the 2nd config with http2 turned on should have failed and used the prior config, verify http2 was not used
tr3.Processes.Default.Streams.stderr = Testers.ExcludesExpression("GET / HTTP/2", "Confirm that HTTP2 is still not used")



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os

Test.Summary = '''
slice plugin prefetch feature test
'''

# Test description:
# Fill origin server with range requests
# Request content through slice plugin with varied prefetch counts

Test.SkipUnless(
    Condition.PluginExists('slice.so'),
    Condition.PluginExists('cache_range_requests.so'),
    Condition.PluginExists('xdebug.so'),
)
Test.ContinueOnFail = False

# configure origin server, lookup by Range header
server = Test.MakeOriginServer("server", lookup_key="{%Range}")

# Define ATS and configure
ts = Test.MakeATSProcess("ts")

block_bytes_1 = 7
block_bytes_2 = 5
body = "lets go surfin now"
bodylen = len(body)

request_header = {
    "headers": "GET /path HTTP/1.1\r\n" + "Host: origin\r\n" + "\r\n",
    "timestamp": "1469733493.993",
    "body": "",
}

response_header = {
    "headers": "HTTP/1.1 200 OK\r\n" + "Connection: close\r\n" + "Cache-Control: public, max-age=5\r\n" + "\r\n",
    "timestamp": "1469733493.993",
    "body": body,
}

server.addResponse("sessionlog.json", request_header, response_header)

# Autest OS doesn't support range request, must manually add requests/responses
for block_bytes in [block_bytes_1, block_bytes_2]:
    for i in range(bodylen // block_bytes + 1):
        b0 = i * block_bytes
        b1 = b0 + block_bytes - 1
        req_header = {
            "headers": "GET /path HTTP/1.1\r\n" + "Host: *\r\n" + "Accept: */*\r\n" + f"Range: bytes={b0}-{b1}\r\n" + "\r\n",
            "timestamp": "1469733493.993",
            "body": ""
        }
        if (b1 > bodylen - 1):
            b1 = bodylen - 1
        resp_header = {
            "headers":
                "HTTP/1.1 206 Partial Content\r\n" + "Accept-Ranges: bytes\r\n" + "Cache-Control: public, max-age=5\r\n" +
                f"Content-Range: bytes {b0}-{b1}/{bodylen}\r\n" + "Connection: close\r\n" + "\r\n",
            "timestamp": "1469733493.993",
            "body": body[b0:b1 + 1]
        }
        server.addResponse("sessionlog.json", req_header, resp_header)

curl_and_args = 'curl -s -D /dev/stdout -o /dev/stderr -x http://127.0.0.1:{} -H "x-debug: x-cache"'.format(ts.Variables.port)

ts.Disk.remap_config.AddLines(
    [
        f'map http://sliceprefetchbytes1/ http://127.0.0.1:{server.Variables.Port}' +
        f' @plugin=slice.so @pparam=--blockbytes-test={block_bytes_1} @pparam=--prefetch-count=1 \\' +
        ' @plugin=cache_range_requests.so',
        f'map http://sliceprefetchbytes2/ http://127.0.0.1:{server.Variables.Port}' +
        f' @plugin=slice.so @pparam=--blockbytes-test={block_bytes_2} @pparam=--prefetch-count=3 \\' +
        ' @plugin=cache_range_requests.so',
    ])

ts.Disk.plugin_config.AddLine('xdebug.so --enable=x-cache')
ts.Disk.logging_yaml.AddLines(
    [
        'logging:',
        '  formats:',
        '  - name: cache',
        '    format: "%<{Content-Range}psh> %<{X-Cache}psh>"',
        '  logs:',
        '    - filename: cache',
        '      format: cache',
        '      mode: ascii',
    ])

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'slice|cache_range_requests|xdebug',
    })

# 0 Test - Full object slice (miss) with only block 14-20 prefetched in background, block bytes= 7
tr = Test.AddTestRun("Full object slice: first block is miss, only block 14-20 prefetched")
ps = tr.Processes.Default
ps.StartBefore(server, ready=When.PortOpen(server.Variables.Port))
ps.StartBefore(Test.Processes.ts)
ps.Command = curl_and_args + ' http://sliceprefetchbytes1/path'
ps.ReturnCode = 0
ps.Streams.stderr = "gold/slice_200.stderr.gold"
ps.Streams.stdout.Content = Testers.ContainsExpression("200 OK", "expected 200 OK response")
ps.Streams.stdout.Content += Testers.ContainsExpression("X-Cache: miss", "expected cache miss")
tr.StillRunningAfter = ts

# 1 Test - Full object slice (hit-fresh) with no prefetched blocks, block bytes= 7
tr = Test.AddTestRun("Full object slice: first block is hit-fresh, no blocks prefetched")
ps = tr.Processes.Default
ps.Command = 'sleep 1; ' + curl_and_args + ' http://sliceprefetchbytes1/path'
ps.ReturnCode = 0
ps.Streams.stderr = "gold/slice_200.stderr.gold"
ps.Streams.stdout.Content = Testers.ContainsExpression("200 OK", "expected 200 OK response")
ps.Streams.stdout.Content += Testers.ContainsExpression("X-Cache: hit-fresh", "expected cache fresh")
tr.StillRunningAfter = ts

# 2 Test - Full object slice with only next block prefetched in background, block bytes= 7
tr = Test.AddTestRun("Full object slice: first block is hit-stale, only block 14-20 prefetched")
ps = tr.Processes.Default
ps.Command = 'sleep 5; ' + curl_and_args + ' http://sliceprefetchbytes1/path'
ps.ReturnCode = 0
ps.Streams.stderr = "gold/slice_200.stderr.gold"
ps.Streams.stdout.Content = Testers.ContainsExpression("200 OK", "expected 200 OK response")
ps.Streams.stdout.Content += Testers.ContainsExpression("X-Cache: hit-stale", "expected cache hit-stale")
tr.StillRunningAfter = ts

# 3 Test - Full object slice (hit-fresh) with no prefetched blocks, block bytes= 7
tr = Test.AddTestRun("Full object slice: first block is hit-fresh with range 0-, no blocks prefetched")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' http://sliceprefetchbytes1/path' + ' -r 0-'
ps.ReturnCode = 0
ps.Streams.stderr = "gold/slice_200.stderr.gold"
ps.Streams.stdout.Content = Testers.ContainsExpression("206 Partial Content", "expected 206 response")
ps.Streams.stdout.Content += Testers.ContainsExpression("Content-Range: bytes 0-17/18", "mismatch byte content response")
ps.Streams.stdout.Content += Testers.ContainsExpression("X-Cache: hit-fresh", "expected cache hit-fresh")
tr.StillRunningAfter = ts

# 4 Test - Client range request (hit-stale/miss) enables prefetching
tr = Test.AddTestRun("Client range request")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' http://sliceprefetchbytes2/path' + ' -r 5-16'
ps.ReturnCode = 0
ps.Streams.stderr = "gold/slice_mid.stderr.gold"
ps.Streams.stdout.Content = Testers.ContainsExpression("206 Partial Content", "expected 206 response")
ps.Streams.stdout.Content += Testers.ContainsExpression("Content-Range: bytes 5-16/18", "mismatch byte content response")
ps.Streams.stdout.Content += Testers.ContainsExpression("X-Cache: miss", "expected cache miss")
tr.StillRunningAfter = ts

# 5 Test - special case, begin inside last slice block but outside asset len
tr = Test.AddTestRun("Invalid end range request, 416")
beg = len(body) + 1
end = beg + block_bytes
ps = tr.Processes.Default
ps.Command = curl_and_args + f' http://sliceprefetchbytes1/path -r {beg}-{end}'
ps.Streams.stdout.Content = Testers.ContainsExpression("416 Requested Range Not Satisfiable", "expected 416 response")
tr.StillRunningAfter = ts

# 6 Test - All requests (client & slice internal) logs to see background fetches
cache_file = os.path.join(ts.Variables.LOGDIR, 'cache.log')
# Wait for log file to appear, then wait one extra second to make sure TS is done writing it.
test_run = Test.AddTestRun("Checking debug logs for background fetches")
test_run.Processes.Default.Command = (os.path.join(Test.Variables.AtsTestToolsDir, 'condwait') + ' 60 1 -f ' + cache_file)
ts.Disk.File(cache_file).Content = "gold/slice_prefetch.gold"
test_run.Processes.Default.ReturnCode = 0
